#!/usr/bin/env python

import argparse
import torch

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='PyTorch RNN/LSTM Language Model')
    parser.add_argument('--vocab', action='store_true', help='print out the full vocab')
    parser.add_argument('load', help='where to load a model from')
    args = parser.parse_args()

    lm = torch.load(args.load, map_location='cpu')
    print(lm.model)
    print("Vocab len:", len(lm.vocab))
    if args.vocab:
        print([c for c in lm.vocab.w2i_])



import torch
from torch import nn
import math
from torch.nn import TransformerEncoder, TransformerEncoderLayer


class PositionalEncoding(nn.Module):
    r"""Inject some information about the relative or absolute position of the tokens
        in the sequence. The positional encodings have the same dimension as
        the embeddings, so that the two can be summed. Here, we use sine and cosine
        functions of different frequencies.
    .. math::
        \text{PosEncoder}(pos, 2i) = sin(pos/10000^(2i/d_model))
        \text{PosEncoder}(pos, 2i+1) = cos(pos/10000^(2i/d_model))
        \text{where pos is the word position and i is the embed idx)
    Args:
        d_model: the embed dim (required).
        dropout: the dropout value (default=0.1).
        max_len: the max. length of the incoming sequence (default=5000).
    Examples:
        >>> pos_encoder = PositionalEncoding(d_model)
    """

    def __init__(self, d_model, dropout=0.1, max_len=5000):
        super(PositionalEncoding, self).__init__()
        self.dropout = nn.Dropout(p=dropout)

        pe = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2).float() * (-math.log(10000.0) / d_model))
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0).transpose(0, 1)
        self.register_buffer('pe', pe)

    def forward(self, x):
        r"""Inputs of forward function
        Args:
            x: the sequence fed to the positional encoder model (required).
        Shape:
            x: [sequence length, batch size, embed dim]
            output: [sequence length, batch size, embed dim]
        Examples:
            >>> output = pos_encoder(x)
        """

        x = x + self.pe[:x.size(0), :]
        return self.dropout(x)


def generate_square_subsequent_mask(sz):
    mask = (torch.triu(torch.ones(sz, sz)) == 1).transpose(0, 1)
    mask = mask.float().masked_fill(mask == 0, float('-inf')).masked_fill(mask == 1, float(0.0))
    return mask


class TransformerLM(nn.Module):
    '''Adapted from PyTorch-examples
    '''
    def __init__(self, vocab_size, nb_heads, dim_res, dim_ff, nb_layers, dropout):
        super().__init__()
        self.src_mask = None
        self.encoder = nn.Embedding(vocab_size, dim_res)
        self.pos_encoder = PositionalEncoding(dim_res, dropout)

        layer_norm = nn.LayerNorm(dim_res)
        encoder_layers = TransformerEncoderLayer(dim_res, nb_heads, dim_ff, dropout)
        self.transformer_encoder = TransformerEncoder(encoder_layers, nb_layers, layer_norm)
        self.dim_res = dim_res

        self.init_weights()
        self.in_len = 1
        self.batch_first = True

    def init_hidden(self, batch_size):
        w = next(self.parameters())
        return torch.zeros((batch_size, 1), dtype=w.dtype, device=w.device)

    def init_weights(self):
        initrange = 0.1
        self.encoder.weight.data.uniform_(-initrange, initrange)

    def forward(self, src, hidden=None):
        src = src.t()
        device = src.device
        if self.src_mask is None or self.src_mask.size(0) != len(src):
            mask = generate_square_subsequent_mask(len(src)).to(device)
            self.src_mask = mask

        src = self.encoder(src) * math.sqrt(self.dim_res)
        src = self.pos_encoder(src)
        output = self.transformer_encoder(src, self.src_mask)

        return output.permute(1, 0, 2), hidden



import numpy as np
import random
import torch


def form_input_targets(stream):
    return stream[:-1], stream[1:]


class CleanStreamsProvider:
    def __init__(self, stream):
        self.stream = stream

    def provide(self):
        return self.stream[:-1], self.stream[1:]


class Corruptor:
    def __init__(self, streams_provider, subs_rate, subs_range, del_rate, ins_rate, protected=[]):
        self.streams_provider = streams_provider
        self.sr = subs_rate
        self.subs_range = subs_range
        self.dr = del_rate
        self.ir = ins_rate
        self.protected = protected

    def provide(self):
        streams = self.streams_provider.provide()
        assert len(streams[0]) == len(streams[1])
        self.inputs = streams[0]
        self.targets = streams[1]

        inputs = []
        targets = []

        i = 0
        nb_nonprotected = 0
        nb_subs = 0
        nb_dels = 0
        nb_inss = 0
        while i < len(self.inputs):
            if self.inputs[i] in self.protected:
                targets.append(self.targets[i])
                inputs.append(self.inputs[i])
                i += 1
                continue

            nb_nonprotected += 1

            roll = random.random()
            if roll < self.dr:
                i += 1
                nb_dels += 1
            elif roll < self.dr + self.sr:
                targets.append(self.targets[i])
                inputs.append(random.randrange(self.subs_range))
                i += 1
                nb_subs += 1
            elif roll < self.dr + self.sr + self.ir:
                targets.append(self.targets[i])
                inputs.append(random.randrange(self.subs_range))
                nb_inss += 1
            else:
                targets.append(self.targets[i])
                inputs.append(self.inputs[i])
                i += 1

        print(f'len {len(self.inputs)}, proper {nb_nonprotected}| D: {100.0*nb_dels/nb_nonprotected:.2f} % ({nb_dels}) S: {100.0*nb_subs/nb_nonprotected:.2f} % ({nb_subs}) I: {100.0*nb_inss/nb_nonprotected:.2f} % ({nb_inss})')
        return torch.tensor(inputs), torch.tensor(targets)


class InputTargetCorruptor:
    def __init__(self, streams_provider, input_subs_rate, target_subs_rate, subs_range, del_rate, ins_rate, protected=[]):
        self.streams_provider = streams_provider
        self.isr = input_subs_rate
        self.tsr = target_subs_rate
        self.subs_range = subs_range
        self.dr = del_rate
        self.ir = ins_rate
        self.protected = protected

    def provide(self):
        streams = self.streams_provider.provide()
        assert len(streams[0]) == len(streams[1])
        self.inputs = streams[0]
        self.targets = streams[1]

        inputs = []
        targets = []

        i = 0
        nb_nonprotected = 0
        nb_subs = 0
        nb_target_subs = 0
        nb_dels = 0
        nb_inss = 0
        while i < len(self.inputs):
            if self.inputs[i] in self.protected:
                targets.append(self.targets[i])
                inputs.append(self.inputs[i])
                i += 1
                continue

            nb_nonprotected += 1

            roll = random.random()
            if roll < self.dr:
                i += 1
                nb_dels += 1
            elif roll < self.dr + self.isr:
                targets.append(self.targets[i])
                inputs.append(random.randrange(self.subs_range))
                i += 1
                nb_subs += 1
            elif roll < self.dr + self.isr:
                targets.append(self.targets[i])
                inputs.append(random.randrange(self.subs_range))
                i += 1
                nb_subs += 1
            elif roll < self.dr + self.isr + self.ir:
                targets.append(self.targets[i])
                inputs.append(random.randrange(self.subs_range))
                nb_inss += 1
            elif roll < self.dr + self.isr + self.ir + self.tsr:
                targets.append(random.randrange(self.subs_range))
                inputs.append(self.inputs[i])
                i += 1
                nb_target_subs += 1
            else:
                targets.append(self.targets[i])
                inputs.append(self.inputs[i])
                i += 1

        print(f'len {len(self.inputs)}, proper {nb_nonprotected}| D: {100.0*nb_dels/nb_nonprotected:.2f} % ({nb_dels}) S: {100.0*nb_subs/nb_nonprotected:.2f} % S_t: {100.0*nb_target_subs/nb_nonprotected:.2f} % ({nb_subs}) I: {100.0*nb_inss/nb_nonprotected:.2f} % ({nb_inss})')
        return torch.tensor(inputs), torch.tensor(targets)


class TargetCorruptor:
    def __init__(self, streams_provider, subs_rate, subs_range, del_rate, ins_rate, protected=[]):
        self.streams_provider = streams_provider
        self.sr = subs_rate
        self.subs_range = subs_range
        self.dr = del_rate
        self.ir = ins_rate
        self.protected = protected

    def provide(self):
        streams = self.streams_provider.provide()
        assert len(streams[0]) == len(streams[1])
        self.inputs = streams[0]
        self.targets = streams[1]

        inputs = []
        targets = []

        i = 0
        nb_nonprotected = 0
        nb_subs = 0
        nb_dels = 0
        nb_inss = 0
        while i < len(self.inputs):
            if self.targets[i] in self.protected:
                targets.append(self.targets[i])
                inputs.append(self.inputs[i])
                i += 1
                continue

            nb_nonprotected += 1

            roll = random.random()
            if roll < self.dr:
                i += 1
                nb_dels += 1
            elif roll < self.dr + self.sr:
                targets.append(random.randrange(self.subs_range))
                inputs.append(self.inputs[i])
                i += 1
                nb_subs += 1
            elif roll < self.dr + self.sr + self.ir:
                targets.append(self.targets[i])
                inputs.append(random.randrange(self.subs_range))
                nb_inss += 1
            else:
                targets.append(self.targets[i])
                inputs.append(self.inputs[i])
                i += 1

        print(f'len {len(self.inputs)}, proper {nb_nonprotected}| D: {100.0*nb_dels/nb_nonprotected:.2f} % ({nb_dels}) S: {100.0*nb_subs/nb_nonprotected:.2f} % ({nb_subs}) I: {100.0*nb_inss/nb_nonprotected:.2f} % ({nb_inss})')
        return torch.tensor(inputs), torch.tensor(targets)


def cut_counts(confusions, mincount):
    counts = {ref_word: {k: v for k, v in nums.items() if v >= mincount} for ref_word, nums in confusions.items()}
    total_counts = {ref_word: sum(counts[ref_word].values()) for ref_word in counts}

    return {ref_word: counts for ref_word, counts in confusions.items() if total_counts[ref_word] > 0}


class Sampler:
    def __init__(self, replacements, probs):
        self.replacements = replacements
        self.probs = probs

    def __call__(self, id, size):
        return np.random.choice(self.replacements[id], size, p=self.probs[id])


class SampleCache:
    def __init__(self, ids, replacements, probs, size=1000):
        self.replacements = replacements
        self.size = size
        self.probs = probs
        self.sampler = Sampler(replacements, probs)
        self.cache = {id: self.sampler(id, size) for id in ids}
        self.backup_cache = {id: self.sampler(id, size) for id in ids}
        self.next = {id: 0 for id in ids}

    def get_next(self, id):
        if self.next[id] == len(self.cache[id]):
            self.cache[id] = self.backup_cache[id]
            self.backup_cache[id] = self.sampler(id, self.size)
            self.next[id] = 0

        sample = self.cache[id][self.next[id]]
        self.next[id] += 1
        return sample


class Confuser:
    NONE_ID = -1

    def __init__(self, confusion_counts, vocab, mincount):
        probs = cut_counts(confusion_counts, mincount)

        def translate(word):
            if word is None:
                return Confuser.NONE_ID
            else:
                return vocab[word]

        id_counts = {translate(ref_word): {translate(replacement): v for replacement, v in stats.items()} for ref_word, stats in probs.items()}
        id_counts_np = {ref_id: (np.asarray(list(probs.keys())), np.asarray(list(probs.values()))) for ref_id, probs in id_counts.items()}

        self.samples_cache = SampleCache(
            list(id_counts_np.keys()),
            {id: repls for id, (repls, _) in id_counts_np.items()},
            {id: counts/np.sum(counts) for id, (_, counts) in id_counts_np.items()},
            size=1000,
        )

    def replace(self, token_id):
        if token_id is None:
            token_id = Confuser.NONE_ID

        try:
            sample = self.samples_cache.get_next(token_id)
        except KeyError:  # New token, never seen in stats. Typical for <unk> and other unks
            return token_id

        if sample == Confuser.NONE_ID:
            return None
        else:
            return sample


class StatisticsCorruptor:
    def __init__(self, streams_provider, confuser, ins_rate, protected=[]):
        self.streams_provider = streams_provider
        self.confuser = confuser
        self.ir = ins_rate
        self.protected = protected

    def provide(self):
        streams = self.streams_provider.provide()
        assert len(streams[0]) == len(streams[1])
        self.inputs = streams[0]
        self.targets = streams[1]

        inputs = []
        targets = []

        i = 0
        nb_nonprotected = 0
        nb_subs = 0
        nb_dels = 0
        nb_inss = 0

        in_len = len(self.inputs)
        while i < in_len:
            in_token = self.inputs[i]
            target_token = self.targets[i]
            if in_token in self.protected:
                targets.append(target_token)
                inputs.append(in_token)
                i += 1
                continue

            nb_nonprotected += 1

            roll = random.random()
            if roll < self.ir:  # Insertion
                insertion = self.confuser.replace(None)
                targets.append(target_token)
                inputs.append(insertion)
                nb_inss += 1
            else:
                repl = self.confuser.replace(in_token)
                if repl == in_token:  # Correct token
                    targets.append(target_token)
                    inputs.append(in_token)
                    i += 1
                elif repl is None:  # Deletion
                    i += 1
                    nb_dels += 1
                else:  # Substitution
                    targets.append(target_token)
                    inputs.append(repl)
                    i += 1
                    nb_subs += 1

        print(f'len {len(self.inputs)}, proper {nb_nonprotected}| D: {100.0*nb_dels/nb_nonprotected:.2f} % ({nb_dels}) S: {100.0*nb_subs/nb_nonprotected:.2f} % ({nb_subs}) I: {100.0*nb_inss/nb_nonprotected:.2f} % ({nb_inss})')
        return torch.tensor(inputs), torch.tensor(targets)


class LazyBatcher:
    def __init__(self, bsz, source):
        self.source = source
        self.bsz = bsz

    def provide(self):
        inputs_stream, targets_stream = self.source.provide()
        len(inputs_stream) == len(targets_stream)

        nb_batches = len(targets_stream) // self.bsz

        inputs_stream = inputs_stream.narrow(0, 0, nb_batches * self.bsz)
        targets_stream = targets_stream.narrow(0, 0, nb_batches * self.bsz)

        self.input_batches = inputs_stream.view(self.bsz, -1).t().contiguous()
        self.target_batches = targets_stream.view(self.bsz, -1).t().contiguous()
        return self.input_batches, self.target_batches


class TemplSplitterClean:
    def __init__(self, target_seq_len, batched_data_provider):
        self.data_provider = batched_data_provider
        self.tsl = target_seq_len

    def __iter__(self):
        data = self.data_provider.provide()
        i = 0
        while i < len(data[0]):
            yield (
                data[0][i:i+self.tsl],
                data[1][i:i+self.tsl],
            )
            i += self.tsl



import torch
import pickle
import zipfile

from brnolm.data_pipeline.masked import masked_tensor_from_sentences


# these helper functions are being developed against an LSTM implementation
# It's ugly, but necessary for Chime. To be made nice later :-)

def detach_hidden_state(h):
    if isinstance(h, tuple):
        return tuple(detach_hidden_state(x) for x in h)
    elif isinstance(h, torch.Tensor):
        return h.detach()
    else:
        return TypeError('Cannot process hidden state represented as {type(h)}')


def split_batch_hidden_state(hidden_state):
    h, c = hidden_state
    return [(h[:, i, :], c[:, i, :]) for i in range(h.shape[1])]


class LanguageModel(torch.nn.Module):
    def __init__(self, model, decoder, vocab):
        super().__init__()

        self.model = model
        self.decoder = decoder
        self.vocab = vocab

    def forward(self, x, h0):
        o, h = self.model(x, h0)
        return self.decoder(o), h

    @property
    def device(self):
        return next(self.parameters()).device

    def single_sentence_nll(self, sentence, prefix):
        '''Provides the negative log-probability of a sequence of tokens
        '''
        sentence_ids = [self.vocab[c] for c in sentence]
        device = self.device

        if prefix:
            prefix_id = self.vocab[prefix]
            tensor = torch.tensor([prefix_id] + sentence_ids).view(1, -1).to(device)
        else:
            tensor = torch.tensor(sentence_ids).view(1, -1).to(device)

        h0 = self.model.init_hidden(1)
        o, _ = self.model(tensor[:, :-1], h0)

        if prefix:
            nll, _ = self.decoder.neg_log_prob(o, tensor[:, 1:])
        else:
            o0 = self.model.extract_output_from_h(h0).unsqueeze(1)
            prepended_o = torch.cat([o0, o], dim=1)
            nll, _ = self.decoder.neg_log_prob(prepended_o, tensor)

        return nll.item()

    def batch_nll(self, sentences, prefix):
        '''Provides the negative log-probability of a batch of sequences of tokens
        '''
        if not sentences:
            return []

        idx_seqs = [[self.vocab[w] for w in s] for s in sentences]

        h0_provider = self.get_custom_h0_provider(prefix)
        masked_nlllh = self.batch_nll_idxs(idx_seqs, h0_provider=h0_provider)

        return masked_nlllh.sum(dim=1).detach().cpu().numpy().tolist()

    def batch_nll_with_h(self, sentences, prefix):
        '''Provides the negative log-probability of a batch of sequences of tokens
        '''
        if not sentences:
            return [], []

        idx_seqs = [[self.vocab[w] for w in s] for s in sentences]

        h0_provider = self.get_custom_h0_provider(prefix)
        masked_nlllh, h = self.batch_nll_idxs(idx_seqs, h0_provider=h0_provider, return_h=True)

        return masked_nlllh.sum(dim=1).detach().cpu().numpy().tolist(), detach_hidden_state(h)

    def batch_nll_idxs(self, idxs, h0_provider=None, return_h=False):
        '''Provides the negative log-probability of a batch of sequences of indexes
        '''
        if h0_provider is None:
            h0_provider = self.model.init_hidden

        device = self.device
        input, target, mask = masked_tensor_from_sentences(idxs, device=device, target_all=True)
        batch_size = input.shape[0]

        h0 = h0_provider(batch_size)
        o, new_h = self.model(input, h0)

        o0 = self.model.extract_output_from_h(h0).unsqueeze(1)
        o = torch.cat([o0, o], dim=1)

        all_nlllh = self.decoder.neg_log_prob_raw(o, target)

        if return_h:
            return all_nlllh * mask, new_h
        else:
            return all_nlllh * mask

    def get_custom_h0_provider(self, prefix):
        if not prefix:
            return self.model.init_hidden

        prefix_ids = [self.vocab[c] for c in prefix]
        device = self.device

        prefix_tensor = torch.tensor(prefix_ids).view(1, -1).to(device)

        def h0_provider(batch_size):
            h0 = self.model.init_hidden(batch_size)
            _, h = self.model(torch.cat([prefix_tensor]*batch_size, axis=0), h0)
            return h

        return h0_provider


def torchscript_export(lm, path):
    s_model = torch.jit.script(lm.model)
    s_dec = torch.jit.script(lm.decoder)
    with zipfile.ZipFile(path, 'w') as zip_f:
        with zip_f.open('decoder.zip', 'w') as dec_f:
            torch.jit.save(s_dec, dec_f)
        with zip_f.open('model.zip', 'w') as model_f:
            torch.jit.save(s_model, model_f)
        with zip_f.open('vocabulary.pickle', 'w') as vocab_f:
            pickle.dump(lm.vocab, vocab_f)


class UnreadableModelError(Exception):
    pass


def torchscript_import(path):
    try:
        with zipfile.ZipFile(path) as zip_f:
            with zip_f.open('decoder.zip') as decoder_f:
                decoder = torch.jit.load(decoder_f)
            with zip_f.open('model.zip') as model_f:
                model = torch.jit.load(model_f)
            with zip_f.open('vocabulary.pickle') as vocab_f:
                vocab = pickle.load(vocab_f)
    except zipfile.BadZipFile as e:
        raise UnreadableModelError(e)

    return LanguageModel(model, decoder, vocab)



class TemporalSplits():
    def __init__(self, seq, nb_inputs_necessary, nb_targets_parallel):
        self._seq = seq
        self._nb_inputs_necessary = nb_inputs_necessary
        self._nb_target_parallel = nb_targets_parallel

    def __iter__(self):
        for lend, rend in self.ranges():
            yield (
                self._seq[lend:rend],
                self._seq[lend+self._nb_inputs_necessary:rend+1]
            )

    def __len__(self):
        return max(len(self._seq) - self._nb_inputs_necessary - self._nb_target_parallel + 1, 0)

    def ranges(self):
        for i in range(0, len(self), self._nb_target_parallel):
            lend = i
            rend = i + self._nb_inputs_necessary + self._nb_target_parallel - 1
            yield lend, rend



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# author : KarelB
# e-mail : ibenes AT fit.vutbr.cz

import argparse
import torch

import .smm_ivec_extractor


def bow_from_sentence(sentence, vocab):
    bow = torch.zeros((1, len(vocab))).float()

    for w in sentence.split():
        bow[0, vocab[w]] += 1.0

    return bow


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--complete-smm", required=True, help="path to a complete SMM model file")
    parser.add_argument('--mkl', default=1, type=int, help='number of MKL threads')

    args = parser.parse_args()

    torch.set_num_threads(args.mkl)
    torch.manual_seed(0)

    with open(args.complete_smm, 'rb') as f:
        ivec_xtractor = smm_ivec_extractor.load(f)

    s_biology1 = "whale rat elephant hippopotamus bee zoology mammals"
    s_biology2 = "flower cat dog insect insect"
    s_buildings1 = "bridge tower house chimney"
    s_buildings2 = "castle factory architecture wall"

    sentences = [s_biology1, s_biology2, s_buildings1, s_buildings2]
    ivecs = [ivec_xtractor(s) for s in sentences]
    ivecs = torch.stack(ivecs)

    print(ivecs @ ivecs.t())



import torch.nn as nn


class FlatEmbedding(nn.Module):
    def __init__(self, nb_tokens, dim_embs, init_range=0.1):
        super().__init__()
        self.embeddings = nn.Embedding(nb_tokens, dim_embs)
        nn.init.uniform_(self.embeddings.weight, -init_range, init_range)

    def forward(self, x):
        return self.embeddings(x)



import logging
import math
import numpy as np
from dataclasses import dataclass

from brnolm.data_pipeline.reading import get_independent_lines, tokens_from_fn
from brnolm.data_pipeline.threaded import OndemandDataProvider
from brnolm.data_pipeline.multistream import Batcher, batchify
from brnolm.data_pipeline.temporal_splitting import TemporalSplits

from brnolm.runtime.runtime_utils import TransposeWrapper
from brnolm.runtime.runtime_multifile import repackage_hidden

from brnolm.data_pipeline.aug_paper_pipeline import Corruptor, form_input_targets, LazyBatcher, TemplSplitterClean
from brnolm.runtime.runtime_utils import CudaStream

import torch


@dataclass
class EvaluationReport:
    total_loss: float
    nb_words: int
    utilization: float

    @property
    def loss_per_token(self):
        return self.total_loss / self.nb_words


class IndependentLinesEvaluator:
    def __init__(self, lm, fn_evalset, max_batch_size, max_tokens, logger=None, total_vocab_size=None):
        if logger:
            self.logger = logger
        else:
            self.logger = logging.getLogger('IndependentLinesEvaluator')
        self.lm = lm
        self.max_batch_size = max_batch_size
        self.max_tokens = max_tokens

        self.logger.debug("preparing data...")
        with open(fn_evalset) as f:
            self.lines = get_independent_lines(f, lm.vocab)

        self.logger.debug("sorting lines...")
        self.lines = sorted(self.lines, key=lambda l: len(l))

        self.logger.debug("computing statistics...")
        self.nb_tokens = sum(len(ids) for ids in self.lines)
        nb_oovs = sum(sum(ids == lm.vocab.unk_ind).detach().item() for ids in self.lines)
        oov_msg = 'Nb oovs: {} / {} ({:.2f} %)\n'.format(nb_oovs, self.nb_tokens, 100.0 * nb_oovs/self.nb_tokens)
        if nb_oovs / self.nb_tokens > 0.05:
            self.logger.warning(oov_msg)
        else:
            self.logger.info(oov_msg)

        self.oov_cost_applicator = OovCostApplicator(
            get_oov_additional_cost(len(self.lm.vocab), total_vocab_size) if total_vocab_size else 0.0,
            lm.vocab.unk_ind,
        )

    def evaluate(self, prefix):
        self.lm.eval()
        h0_provider = self.lm.get_custom_h0_provider(prefix.split())

        loss = 0.0
        data_stream = OndemandDataProvider(Batcher(self.lines, self.max_batch_size, self.max_tokens), device=self.lm.device)
        total_actual_size = 0
        with torch.no_grad():
            for i, batch in enumerate(data_stream):
                per_line_losses = self.lm.batch_nll_idxs(batch, h0_provider)
                for i, (idxs, losses) in enumerate(zip(batch, per_line_losses)):
                    per_line_losses[i] = self.oov_cost_applicator(idxs, losses)
                loss += per_line_losses.sum().detach().item()
                total_actual_size += per_line_losses.numel()

        utilization = self.nb_tokens/total_actual_size
        return EvaluationReport(loss, self.nb_tokens, utilization)


class EnblockEvaluator:
    def __init__(self, lm, data_fn, batch_size, target_seq_len, logger=None, tokenize_regime='words'):
        if logger:
            self.logger = logger
        else:
            self.logger = logging.getLogger('EnblockEvaluator')
        self.batch_size = batch_size
        self.lm = lm

        ids = tokens_from_fn(data_fn, lm.vocab, regime=tokenize_regime, randomize=False)
        oov_mask = ids == lm.vocab.unk_ind
        nb_oovs = oov_mask.sum().item()

        nb_tokens = len(ids)
        oov_msg = 'Nb oovs: {} / {} ({:.2f} %)\n'.format(nb_oovs, len(ids), 100.0 * nb_oovs/nb_tokens)
        if nb_oovs / nb_tokens > 0.05:
            self.logger.warning(oov_msg)
        else:
            self.logger.info(oov_msg)

        batched = batchify(ids, batch_size, lm.device == torch.device('cuda:0'))
        data_tb = TemporalSplits(
            batched,
            nb_inputs_necessary=lm.model.in_len,
            nb_targets_parallel=target_seq_len
        )
        self.data = TransposeWrapper(data_tb)

    def evaluate(self):
        self.lm.eval()

        total_loss = 0.0
        total_timesteps = 0
        hidden = self.lm.model.init_hidden(self.batch_size)

        for X, targets in self.data:
            hidden = repackage_hidden(hidden)

            output, hidden = self.lm.model(X, hidden)
            losses = self.lm.decoder.neg_log_prob_raw(output, targets)

            total_loss += losses.sum().detach()
            total_timesteps += targets.numel()

        return EvaluationReport(total_loss.item(), total_timesteps, 1.0)


class SubstitutionalEnblockEvaluator:
    def __init__(self, lm, data_fn, batch_size, target_seq_len, corruptor, nb_rounds, logger=None, tokenize_regime='words'):
        if logger:
            self.logger = logger
        else:
            self.logger = logging.getLogger('SubstitutionalEnblockEvaluator')
        self.batch_size = batch_size
        self.lm = lm
        self.nb_rounds = nb_rounds

        ids = tokens_from_fn(data_fn, lm.vocab, regime=tokenize_regime, randomize=False)
        oov_mask = ids == lm.vocab.unk_ind
        nb_oovs = oov_mask.sum().item()

        nb_tokens = len(ids)
        oov_msg = 'Nb oovs: {} / {} ({:.2f} %)\n'.format(nb_oovs, len(ids), 100.0 * nb_oovs/nb_tokens)
        if nb_oovs / nb_tokens > 0.05:
            self.logger.warning(oov_msg)
        else:
            self.logger.info(oov_msg)

        batched = batchify(ids, batch_size, lm.device == torch.device('cuda:0'))
        data_tb = TemporalSplits(
            batched,
            nb_inputs_necessary=lm.model.in_len,
            nb_targets_parallel=target_seq_len
        )
        self.data = corruptor(TransposeWrapper(data_tb))

    def evaluate(self, report_individual=False):
        overall_total_loss = 0.0
        overall_total_timesteps = 0.0
        ppls = []

        for round_no in range(self.nb_rounds):
            self.lm.eval()

            total_loss = 0.0
            total_timesteps = 0
            hidden = self.lm.model.init_hidden(self.batch_size)

            for X, targets in self.data:
                hidden = repackage_hidden(hidden)

                output, hidden = self.lm.model(X, hidden)
                losses = self.lm.decoder.neg_log_prob_raw(output, targets)

                total_loss += losses.sum().detach()
                total_timesteps += targets.numel()

            eval_report = EvaluationReport(total_loss.item(), total_timesteps, 1.0)
            ppl = math.exp(eval_report.loss_per_token)
            if report_individual:
                self.logger.info('total loss {:.1f} | per token loss {:5.2f} | ppl {:8.2f}'.format(eval_report.total_loss, eval_report.loss_per_token, ppl))

            overall_total_loss += total_loss
            overall_total_timesteps += total_timesteps
            ppls.append(ppl)

        ppls = np.asarray(ppls)
        self.logger.info(f'PPLs summary: {np.min(ppls):.2f} / {np.mean(ppls):.2f} / {np.max(ppls):.2f} , stddev: {np.std(ppls):.3f}')
        return EvaluationReport(overall_total_loss.item(), overall_total_timesteps, 1.0)


class SubstitutionalEnblockEvaluator_v2:
    def __init__(self, lm, data_fn, batch_size, target_seq_len, corruptor, nb_rounds, logger=None, tokenize_regime='words'):
        if logger:
            self.logger = logger
        else:
            self.logger = logging.getLogger('SubstitutionalEnblockEvaluator_v2')
        self.batch_size = batch_size
        self.lm = lm
        self.nb_rounds = nb_rounds

        ids = tokens_from_fn(data_fn, lm.vocab, regime=tokenize_regime, randomize=False)
        oov_mask = ids == lm.vocab.unk_ind
        nb_oovs = oov_mask.sum().item()

        nb_tokens = len(ids)
        oov_msg = 'Nb oovs: {} / {} ({:.2f} %)\n'.format(nb_oovs, len(ids), 100.0 * nb_oovs/nb_tokens)
        if nb_oovs / nb_tokens > 0.05:
            self.logger.warning(oov_msg)
        else:
            self.logger.info(oov_msg)

        streams = form_input_targets(ids)
        corrupted_provider = corruptor(streams)
        batch_former = LazyBatcher(batch_size, corrupted_provider)
        data_tb = TemplSplitterClean(target_seq_len, batch_former)

        self.data = CudaStream(TransposeWrapper(data_tb))

    def evaluate(self, report_individual=False):
        overall_total_loss = 0.0
        overall_total_timesteps = 0.0
        ppls = []

        for round_no in range(self.nb_rounds):
            self.lm.eval()

            total_loss = 0.0
            total_timesteps = 0
            hidden = self.lm.model.init_hidden(self.batch_size)

            for X, targets in self.data:
                hidden = repackage_hidden(hidden)

                output, hidden = self.lm.model(X, hidden)
                losses = self.lm.decoder.neg_log_prob_raw(output, targets)

                total_loss += losses.sum().detach()
                total_timesteps += targets.numel()

            eval_report = EvaluationReport(total_loss.item(), total_timesteps, 1.0)
            ppl = math.exp(eval_report.loss_per_token)
            if report_individual:
                self.logger.info('total loss {:.1f} | per token loss {:5.2f} | ppl {:8.2f}'.format(eval_report.total_loss, eval_report.loss_per_token, ppl))

            overall_total_loss += total_loss
            overall_total_timesteps += total_timesteps
            ppls.append(ppl)

        ppls = np.asarray(ppls)
        self.logger.info(f'PPLs summary: {np.min(ppls):.2f} / {np.mean(ppls):.2f} / {np.max(ppls):.2f} , stddev: {np.std(ppls):.3f}')
        return EvaluationReport(overall_total_loss.item(), overall_total_timesteps, 1.0)


def get_oov_additional_cost(lm_vocab_size, total_vocab_size):
    nb_oovs_uncovered = total_vocab_size - lm_vocab_size
    return math.log(nb_oovs_uncovered)


class OovCostApplicator:
    def __init__(self, oov_penalty, unk_ind):
        self.oov_penalty = oov_penalty
        self.unk_ind = unk_ind

    def __call__(self, ids, losses):
        if self.oov_penalty == 0.0:
            return losses

        unk_mask = (ids == self.unk_ind).to(losses.device)
        zero_padding = torch.zeros(
            (len(losses) - len(ids),),
            dtype=unk_mask.dtype, device=unk_mask.device
        )
        unk_mask = torch.cat([unk_mask, zero_padding])

        return losses + unk_mask*self.oov_penalty



import sys
import time
import math

import torch


class BaseLogger():
    def __init__(self, report_period, output_file=sys.stdout):
        self._start_time = time.time()
        self._nb_logs = 0
        self._report_period = report_period
        self._of = output_file
        self._construction_time = time.time()

    def log(self, *args):
        self._log(*args)
        self._nb_logs += 1

        if self._nb_logs % self._report_period == 0:
            self._flush()
            self._reset()
            self._start_time = time.time()

    def nb_updates(self):
        return self._nb_logs

    def time_since_creation(self):
        return time.time() - self._construction_time

    def _flush(self):
        pass

    def _reset(self):
        pass

    def _log(self, *args):
        pass


class InfinityLogger(BaseLogger):
    def __init__(self, epoch, report_period, lr, output_file=sys.stdout):
        super().__init__(report_period, output_file)
        self._running_loss = 0.0
        self._epoch = epoch
        self._lr = lr

    def _log(self, loss):
        self._running_loss += loss

    def _flush(self):
        ms_per_log = (time.time() - self._start_time) * 1000 / self._report_period
        cur_loss = (self._running_loss / self._report_period).item()
        fmt_string = '| epoch {:3d} | {:5d} batches done | lr {:.3e} | ms/batch {:5.2f} | loss {:5.2f} | ppl {:8.2f}\n'
        line = fmt_string.format(
            self._epoch, self._nb_logs, self._lr,
            ms_per_log, cur_loss, math.exp(cur_loss)
        )
        self._of.write(line)

    def _reset(self):
        self._running_loss = 0.0


class GradLogger(BaseLogger):
    def __init__(self, report_period, named_params, output_file=sys.stdout):
        super().__init__(report_period, output_file)
        self._named_params = list(named_params)

        self._grads = {}
        for name, param in self._named_params:
            self._grads[name] = []

        self._of.write("{}\n".format(" ".join(self._grads)))

    def _log(self):
        for name, param in self._named_params:
            self._grads[name].append(param.grad.abs().mean())

    def _flush(self):
        grad_mavs = []
        for name in self._grads:
            all_grads = torch.stack(self._grads[name])
            grad_mavs.append(all_grads.mean().data.item())

        fmt_string = " ".join("{:.7f}" for _ in grad_mavs) + "\n"
        line = fmt_string.format(*grad_mavs)
        self._of.write(line)

    def _reset(self):
        for name in self._grads:
            self._grads[name] = []


class ProgressLogger():
    def __init__(self, epoch, report_period, lr, nb_updates, output_file=sys.stdout):
        self._start_time = time.time()
        self._nb_logs = 0
        self._running_loss = 0.0
        self._epoch = epoch
        self._report_period = report_period
        self._of = output_file
        self._lr = lr
        self._construction_time = time.time()
        self._nb_updates = nb_updates

    def log(self, loss):
        self._running_loss += loss
        self._nb_logs += 1

        if self._nb_logs % self._report_period == 0:
            self._flush()
            self._reset()

    def time_since_creation(self):
        return time.time() - self._construction_time

    def nb_updates(self):
        return self._nb_updates

    def _flush(self):
        ms_per_log = (time.time() - self._start_time) * 1000 / self._report_period
        cur_loss = (self._running_loss / self._report_period).item()
        fmt_string = '| epoch {:3d} | {:5d}/{:5d} batches | lr {:.3e} | ms/batch {:5.2f} | loss {:5.2f} | ppl {:8.2f}\n'
        line = fmt_string.format(
            self._epoch, self._nb_logs, self._nb_updates, self._lr,
            ms_per_log, cur_loss, math.exp(cur_loss)
        )
        self._of.write(line)

    def _reset(self):
        self._running_loss = 0.0
        self._start_time = time.time()


class NoneLogger():
    def log(self, *args):
        pass



import torch


class FullSoftmaxDecoder(torch.nn.Module):
    def __init__(self, nb_hidden, nb_output, init_range=0.1):
        super().__init__()

        self.projection = torch.nn.Linear(nb_hidden, nb_output)
        self.log_softmax = torch.nn.LogSoftmax(dim=-1)

        self.projection.weight.data.uniform_(-init_range, init_range)
        self.projection.bias.data.fill_(0)

        self.nllloss = torch.nn.NLLLoss(reduction='sum')

    def forward(self, X):
        a = self.projection(X)
        return self.log_softmax(a)

    def neg_log_prob_raw(self, X, targets):
        orig_shape = targets.shape
        preds = self.forward(X)
        targets_flat = targets.view(-1)
        preds_flat = preds.view(-1, preds.size(-1))

        return torch.nn.functional.nll_loss(preds_flat, targets_flat, reduction='none').view(orig_shape)

    @torch.jit.export
    def neg_log_prob(self, X, targets):
        return self.neg_log_prob_raw(X, targets).sum(), targets.numel()


class CustomLossFullSoftmaxDecoder(torch.nn.Module):
    def __init__(self, nb_hidden, nb_output, init_range=0.1, label_smoothing=None):
        super().__init__()

        assert label_smoothing is None or (label_smoothing >= 0.0 and label_smoothing < 1.0)

        self.projection = torch.nn.Linear(nb_hidden, nb_output)
        self.log_softmax = torch.nn.LogSoftmax(dim=-1)

        self.projection.weight.data.uniform_(-init_range, init_range)
        self.projection.bias.data.fill_(0)

        if label_smoothing is not None:
            self.core_loss = LabelSmoothedNLLLoss(label_smoothing)
        else:
            self.core_loss = plain_nll_loss

    def forward(self, X):
        a = self.projection(X)
        return self.log_softmax(a)

    def neg_log_prob_raw(self, X, targets):
        orig_shape = targets.shape
        preds = self.forward(X)
        targets_flat = targets.view(-1)
        preds_flat = preds.view(-1, preds.size(-1))

        return self.core_loss(preds_flat, targets_flat).view(orig_shape)

    def neg_log_prob(self, X, targets):
        return self.neg_log_prob_raw(X, targets).sum(), targets.numel()


def plain_nll_loss(preds, targets):
    return torch.nn.functional.nll_loss(preds, targets, reduction='none')


class LabelSmoothedNLLLoss:
    def __init__(self, amount):
        self.amount = amount

    def __call__(self, preds, targets):
        eps = self.amount
        n_class = preds.size(1)

        assert len(targets.shape) == 1
        assert len(preds.shape) == 2

        smooth_targets = torch.zeros_like(preds).scatter_(1, targets.unsqueeze(1), 1)
        smooth_targets = smooth_targets * (1 - eps) + (1 - smooth_targets) * eps / n_class
        loss = -(smooth_targets * preds).sum(axis=1)

        return loss



#!/usr/bin/env python
 
import argparse

import lstm_model
import vocab
import language_model

import scipy.io as sio
from sklearn.feature_extraction.text import CountVectorizer

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='PyTorch LSTM Language Model')
    parser.add_argument('--filelist', type=str,  required=True,
                        help='list of files with documents (one per line)')
    parser.add_argument('--load', type=str,  required=True,
                        help='path to corresponding NN LM')
    parser.add_argument('--save', type=str,  required=True,
                        help='path to mtx file')
    args = parser.parse_args()

    print("loading model...")
    with open(args.load, 'rb') as f:
        lm = language_model.load(f)
    print(lm.model)
    vocab = lm.vocab

    documents = []
    with open(args.filelist) as fl:
        filenames = fl.read().split()
        for filename in filenames:
            with open(filename) as f:
                documents.append(f.read())
         
    cvect = CountVectorizer(documents, analyzer='word', lowercase=False, vocabulary=vocab)

    document_bows = cvect.fit_transform(documents)
    vocab = cvect.get_feature_names()
    print('document_bows:', document_bows.shape)

    sio.mmwrite(args.save, document_bows.T)



def scaled_int_str(value):
    if value < 1000:
        return f'{value}'
    elif value < 1000000:
        return f'{value/1000:.1f}k'
    else:
        return f'{value/1000000:.1f}M'


class ModelStatistics:
    def __init__(self, model):
        self.model = model

    def total_nb_params(self):
        return sum(p.numel() for p in self.model.parameters())

    def nb_trainable_params(self):
        return sum(p.numel() for p in self.model.parameters() if p.requires_grad)

    def trainable_params_breakup(self):
        per_param_desc = (f'{name} {scaled_int_str(p.numel())}\n' for name, p in self.model.named_parameters() if p.requires_grad)
        return ''.join(per_param_desc)

    def nb_nontrainable_params(self):
        return sum(p.numel() for p in self.model.parameters() if not p.requires_grad)

    def __str__(self):
        torch_desc = f'{self.model}\n'
        nb_params_desc = f'Total number of parameters: {scaled_int_str(self.total_nb_params())}\n'
        nb_trainable_desc = f'Number of trainable parameters: {scaled_int_str(self.nb_trainable_params())}\n'
        nb_nontrainable_desc = f'Number of nontrainable parameters: {scaled_int_str(self.nb_nontrainable_params())}\n'
        return torch_desc + nb_params_desc + nb_trainable_desc + nb_nontrainable_desc + self.trainable_params_breakup()



#!/usr/bin/env python

import argparse
import math
import runtime_utils
import vocab

import torch
import analysis


def bows_to_ps(bows):
    uni_ps = bows.t() / bows.sum(dim=1)
    return uni_ps.t()
    

def bows_to_ent(bows):
    uni_ps = bows_to_ps(bows)
    entropies = analysis.categorical_entropy(uni_ps)

    avg_entropy = entropies @ bows.sum(dim=1) / bows.sum()
    return avg_entropy

def documents_from_fn(fn_filelist):
    fns = runtime_utils.filenames_file_to_filenames(fn_filelist)
    documents = []
    for fn in fns:
        with open(fn) as f:
            documents.append(f.read().split())

    return documents

def bow_from_documents(documents, vocab):
    bows = torch.zeros(len(documents), len(vocab)).long()
    for doc_no, doc in enumerate(documents):
        for w in doc:
            bows[doc_no, vocab[w]] += 1

    return bows
    

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--source-list')
    parser.add_argument('--file-list')
    parser.add_argument('--vocab')
    parser.add_argument('--unk', default='<unk>')
    args = parser.parse_args()

    with open(args.vocab) as f:
        vocab = vocab.vocab_from_kaldi_wordlist(f, args.unk)

    documents = documents_from_fn(args.source_list)
    bows = bow_from_documents(documents, vocab).float()
    unigram_ps = bows_to_ps(bows.sum(dim=0, keepdim=True)).squeeze()

    test_documents = documents_from_fn(args.file_list)
    test_bows = bow_from_documents(test_documents, vocab).float()
    test_unigrams = bows_to_ps(test_bows)

    print(unigram_ps.size())
    print(test_unigrams.size())

    cross_entropies = analysis.categorical_cross_entropy(test_unigrams, unigram_ps)
    # print(cross_entropies)
    
    test_lengths = test_bows.sum(dim=1)
    avg_entropy =  cross_entropies @ test_lengths / test_bows.sum()
    print("{:.4f} {:.2f}".format(avg_entropy, 2**avg_entropy))



from .common import TestCase
import os
import sys
import unittest

import numpy as np
import torch
from sklearn.feature_extraction.text import CountVectorizer

import brnolm.data_pipeline.split_corpus_dataset as split_corpus_dataset
import brnolm.smm_itf.ivec_appenders as ivec_appenders
from brnolm.language_models.vocab import Vocabulary

from .utils import getStream

try:
    import brnolm.smm_itf.smm_ivec_extractor as smm_ivec_extractor
    from smm import SMM, estimate_ubm
except ImportError:
    sys.stderr.write('Failed to import SMM implementation\n')


@unittest.skipIf(os.environ.get('TEST_SMM') != 'yes', "For SMM tests, set TEST_SMM='yes'")
class CheatingIvecAppenderTests(TestCase):
    def setUp(self):
        self.ivec_eetor = lambda x: np.asarray([hash(x) % 1337])
        self.test_words_short = "a b c a".split()
        self.test_words_long = "a b c a a".split()
        self.vocab = {
            "a": 0,
            "b": 1,
            "c": 2
        }

    def test_single_data(self):
        data_source = getStream(self.test_words_short)
        ts = split_corpus_dataset.TokenizedSplit(data_source, self.vocab, 1)
        appender = ivec_appenders.CheatingIvecAppender(ts, self.ivec_eetor)

        # cannot acces ts._tokens, it's an implementation
        tokens = [self.vocab[w] for w in self.test_words_short]

        expectation = self.ivec_eetor(" ".join(self.test_words_short[:-1]))
        seqs = next(iter(appender))
        first = seqs[2]

        self.assertEqual(first, expectation)

    def test_whole_seq(self):
        data_source = getStream(self.test_words_short)
        ts = split_corpus_dataset.TokenizedSplit(data_source, self.vocab, 1)
        appender = ivec_appenders.CheatingIvecAppender(ts, self.ivec_eetor)

        # cannot acces ts._tokens, it's an implementation
        tokens = [self.vocab[w] for w in self.test_words_short]

        expectation = [
            self.ivec_eetor(" ".join(self.test_words_short[:-1])),
            self.ivec_eetor(" ".join(self.test_words_short[:-1])),
            self.ivec_eetor(" ".join(self.test_words_short[:-1]))
        ]

        seqs = [x[2] for x in (iter(appender))]
        self.assertEqual(seqs, expectation)

    def test_whole_seq_with_next(self):
        data_source = getStream(self.test_words_short)
        ts = split_corpus_dataset.TokenizedSplit(data_source, self.vocab, 1)
        appender = ivec_appenders.CheatingIvecAppender(ts, self.ivec_eetor)
        appender = iter(appender)

        # cannot acces ts._tokens, it's an implementation
        tokens = [self.vocab[w] for w in self.test_words_short]
        expectation = [
            self.ivec_eetor(" ".join(self.test_words_short[:-1])),
            self.ivec_eetor(" ".join(self.test_words_short[:-1])),
            self.ivec_eetor(" ".join(self.test_words_short[:-1]))
        ]

        seq0 = next(appender)[2]
        self.assertEqual(seq0, expectation[0])

        seq1 = next(appender)[2]
        self.assertEqual(seq1, expectation[1])

        seq2 = next(appender)[2]
        self.assertEqual(seq2, expectation[2])

    def test_iter_ends(self):
        data_source = getStream(self.test_words_short)
        ts = split_corpus_dataset.TokenizedSplit(data_source, self.vocab, 1)
        appender = ivec_appenders.CheatingIvecAppender(ts, self.ivec_eetor)
        appender = iter(appender)

        next(appender)
        next(appender)
        next(appender)

        self.assertRaises(StopIteration, next, appender)


@unittest.skipIf(os.environ.get('TEST_SMM') != 'yes', "For SMM tests, set TEST_SMM='yes'")
class HistoryIvecAppenderTests(TestCase):
    def setUp(self):
        self.ivec_eetor = lambda x: np.asarray([hash(x) % 1337])
        self.test_words_short = "a b c a".split()
        self.test_words_long = "ab bb cd a b".split()
        self.vocab = {
            "a": 0,
            "b": 1,
            "c": 2,
            "ab": 3,
            "bb": 4,
            "cd": 5,
        }

    def test_single_data(self):
        data_source = getStream(self.test_words_short)
        ts = split_corpus_dataset.TokenizedSplit(data_source, self.vocab, 1)
        appender = ivec_appenders.HistoryIvecAppender(ts, self.ivec_eetor)

        # cannot acces ts._tokens, it's an implementation
        tokens = [self.vocab[w] for w in self.test_words_short]

        expectation = self.ivec_eetor(" ".join([]))
        seqs = next(iter(appender))
        first = seqs[2]

        self.assertEqual(first, expectation)

    def test_whole_seq(self):
        data_source = getStream(self.test_words_short)
        ts = split_corpus_dataset.TokenizedSplit(data_source, self.vocab, 1)
        appender = ivec_appenders.HistoryIvecAppender(ts, self.ivec_eetor)

        # cannot acces ts._tokens, it's an implementation
        tokens = [self.vocab[w] for w in self.test_words_short]

        expectation = [
            self.ivec_eetor(" ".join(self.test_words_short[:0])),
            self.ivec_eetor(" ".join(self.test_words_short[:1])),
            self.ivec_eetor(" ".join(self.test_words_short[:2])),
        ]
        seqs = [x[2] for x in (iter(appender))]

        self.assertEqual(seqs, expectation)

    def test_multiletter_words(self):
        data_source = getStream(self.test_words_long)
        ts = split_corpus_dataset.TokenizedSplit(data_source, self.vocab, 1)
        appender = ivec_appenders.HistoryIvecAppender(ts, self.ivec_eetor)

        # cannot acces ts._tokens, it's an implementation
        tokens = [self.vocab[w] for w in self.test_words_short]

        expectation = [
            self.ivec_eetor(" ".join(self.test_words_long[:0])),
            self.ivec_eetor(" ".join(self.test_words_long[:1])),
            self.ivec_eetor(" ".join(self.test_words_long[:2])),
            self.ivec_eetor(" ".join(self.test_words_long[:3])),
        ]
        seqs = [x[2] for x in iter(appender)]

        self.assertEqual(seqs, expectation)


@unittest.skipIf(os.environ.get('TEST_SMM') != 'yes', "For SMM tests, set TEST_SMM='yes'")
class ParalelIvecAppenderTests(TestCase):
    def setUp(self):
        documents = ["a set of documents containing eigth different", "words of documents"]
        cvect = CountVectorizer(documents, strip_accents='ascii', analyzer='word')
        cvect.fit(documents)
        vocab = cvect.get_feature_names()
        self.cvect = CountVectorizer(documents, strip_accents='ascii', analyzer='word', vocabulary=vocab)

        bows = cvect.transform(documents)
        bows = bows.A.astype(np.float32).T

        ubm = estimate_ubm(bows)
        hyper = {
            'lam_w': 0.1,
            'reg_t': 0.1,
            'lam_t': 0.1,
            'iv_dim': 4
        }
        smm = SMM(bows.shape[0], ubm, hyper)

        self.extractor = smm_ivec_extractor.IvecExtractor(smm, nb_iters=10, lr=0.1, tokenizer=self.cvect)

        self.vocab = Vocabulary(unk_word="<unk>", unk_index=0)
        self.vocab.add_from_text(" ".join(documents))

        self.translator = self.extractor.build_translator(self.vocab)

    def test_single_data(self):
        ws = [self.vocab[w] for w in self.vocab]

        xs = [
            torch.LongTensor([[ws[0], ws[1]], [ws[0], ws[2]]]),
            torch.LongTensor([[ws[2], ws[3]], [ws[4], ws[6]]]),
            torch.LongTensor([[ws[4], ws[5]], [ws[0], ws[1]]]),
            torch.LongTensor([[ws[2], ws[3]], [ws[0], ws[4]]]),
        ]

        ts = [
            torch.LongTensor([[ws[1], ws[2]], [ws[1], ws[3]]]),
            torch.LongTensor([[ws[3], ws[4]], [ws[5], ws[7]]]),
            torch.LongTensor([[ws[5], ws[6]], [ws[1], ws[2]]]),
            torch.LongTensor([[ws[3], ws[4]], [ws[1], ws[5]]]),
        ]

        masks = [
            torch.LongTensor([]),
            torch.LongTensor([0, 1]),
            torch.LongTensor([0]),
            torch.LongTensor([1]),
        ]

        stream = zip(xs, ts, masks)
        ivec_appender = ivec_appenders.ParalelIvecAppender(stream, self.extractor, self.translator)

        exp_ivecs = [
            self.extractor(self.extractor.zero_bows(2)),
            self.extractor(self.translator(xs[0])),
            self.extractor(torch.stack([
                self.translator(torch.cat([xs[0][0], xs[1][0]])),
                self.extractor.zero_bows(1).squeeze()
            ])),
            self.extractor(torch.stack([
                self.translator(xs[2][1]),
                self.extractor.zero_bows(1).squeeze(),
            ])),
        ]

        exp_stream = list(zip(xs, ts, exp_ivecs, masks))
        obs_stream = list(ivec_appender)

        self.assertEqual(obs_stream[0], exp_stream[0])
        self.assertEqual(obs_stream[1], exp_stream[1])
        self.assertEqual(obs_stream[2], exp_stream[2])
        self.assertEqual(obs_stream[3], exp_stream[3])



import unittest
import torch 

from brnolm.analysis import categorical_entropy, categorical_cross_entropy, categorical_kld


class CategoricalEntropyTests(unittest.TestCase):
    def setUp(self):
        pass


    def test_uniform_2d(self):
        p_x = torch.FloatTensor([0.5, 0.5])
        H_x = torch.FloatTensor([1.0])

        self.assertTrue(torch.equal(categorical_entropy(p_x), H_x))


    def test_uniform_4d(self):
        p_x = torch.FloatTensor([0.25, 0.25, 0.25, 0.25])
        H_x = torch.FloatTensor([2.0])

        self.assertTrue(torch.equal(categorical_entropy(p_x), H_x))


    def test_nonuniform_3d(self):
        p_x = torch.FloatTensor([0.5, 0.25, 0.25])
        H_x = torch.FloatTensor([1.5])

        self.assertTrue(torch.equal(categorical_entropy(p_x), H_x))


    def test_sparse_2d(self):
        p_x = torch.FloatTensor([0.0, 1.0])
        H_x = torch.FloatTensor([0.0])

        H_x_hat = categorical_entropy(p_x)

        self.assertTrue(torch.equal(H_x_hat, H_x))


    def test_uniform_2_dists(self):
        p_x = torch.FloatTensor([[0.5, 0.5], [0.0, 1.0]])
        H_x = torch.FloatTensor([1.0, 0.0])

        H_x_hat = categorical_entropy(p_x) 

        self.assertTrue(torch.equal(H_x_hat, H_x))


class CategoricalCrossEntropyTests(unittest.TestCase):
    def setUp(self):
        pass


    def test_simple(self):
        p_x = torch.FloatTensor([0.5, 0.5])
        q_x = torch.FloatTensor([0.25, 0.75])

        xent = categorical_cross_entropy(p_x, q_x)

        self.assertAlmostEqual(xent[0], 1.207518749639422, delta=1e-7)


    def test_sparse_true_dist(self):
        p_x = torch.FloatTensor([1, 0.0])
        q_x = torch.FloatTensor([0.25, 0.75])

        xent = categorical_cross_entropy(p_x, q_x)

        self.assertEqual(xent[0], 2)


    def test_sparse_both_same(self):
        p_x = torch.FloatTensor([0.5, 0.5, 0.0])
        q_x = torch.FloatTensor([0.5, 0.5, 0.0])

        xent = categorical_cross_entropy(p_x, q_x)

        self.assertEqual(xent[0], 1)


    def test_sparse_both_different(self):
        p_x = torch.FloatTensor([0.5, 0.0, 0.5])
        q_x = torch.FloatTensor([0.5, 0.5, 0.0])

        xent = categorical_cross_entropy(p_x, q_x)

        self.assertEqual(xent[0], float("inf"))


    def test_2_dists(self):
        p_x = torch.FloatTensor([[0.5, 0.5], [1.0, 0.0]])
        q_x = torch.FloatTensor([[0.25, 0.75], [0.25, 0.75]])

        xent = categorical_cross_entropy(p_x, q_x)

        self.assertAlmostEqual(xent[0], 1.207518749639422, delta=1e-7)
        self.assertEqual(xent[1], 2)


    def test_one_vs_many(self):
        p_x = torch.FloatTensor([[0.5, 0.5], [1.0, 0.0]])
        q_x = torch.FloatTensor([0.25, 0.75])

        xent = categorical_cross_entropy(p_x, q_x)

        self.assertAlmostEqual(xent[0], 1.207518749639422, delta=1e-7)
        self.assertEqual(xent[1], 2)


class CategoricalKLDTests(unittest.TestCase):
    def setUp(self):
        pass


    def test_same(self):
        p_x = torch.FloatTensor([0.5, 0.5])
        q_x = torch.FloatTensor([0.5, 0.5])

        kld = categorical_kld(p_x, q_x)

        self.assertAlmostEqual(kld[0], 0.0)


    def test_simple_different(self):
        p_x = torch.FloatTensor([0.5, 0.5])
        q_x = torch.FloatTensor([0.75, 0.25])

        kld = categorical_kld(p_x, q_x)

        self.assertAlmostEqual(kld[0], 0.207518749639422, delta=1e-7)


    def test_true_has_zero(self):
        p_x = torch.FloatTensor([1.0, 0.0])
        q_x = torch.FloatTensor([0.25, 0.75])

        kld = categorical_kld(p_x, q_x)

        self.assertAlmostEqual(kld[0], 2, delta=1e-7)


    def test_infinite_kld(self):
        p_x = torch.FloatTensor([0.5, 0.5])
        q_x = torch.FloatTensor([0.0, 1.0])

        kld = categorical_kld(p_x, q_x)

        self.assertEqual(kld[0], float("inf"))


    def test_2_dists(self):
        p_x = torch.FloatTensor([[0.5, 0.5], [0.5, 0.5]])
        q_x = torch.FloatTensor([[0.0, 1.0], [0.5, 0.5]])

        kld = categorical_kld(p_x, q_x)

        self.assertEqual(kld[0], float("inf"))
        self.assertEqual(kld[1], 0.0)
        



#!/usr/bin/env python
import argparse
import io
import math
import sys
import torch

import split_corpus_dataset
import ivec_appenders
import smm_ivec_extractor

from runtime_utils import filenames_file_to_filenames

class DummyDict:
    def __getitem__(self, index):
        return 0


def euclidean_distance(a, b):
    return (a-b).pow(2).sum(dim=-1).pow(0.5)


def length(a):
    return a.pow(2).sum(dim=-1).pow(0.5)


def euclidean_distance(a, b):
    return (a-b).pow(2).sum(dim=-1).pow(0.5)


def cosine_similarity(a, b):
    return (a*b).sum(dim=-1) / (length(a) * length(b))


def analyze_document(text, ivec_extractor):
    nb_words = len(text.split())

    if args.unroll_steps is None:
        unroll = args.unroll
    else:
        if nb_words % args.unroll_steps == 0:
            text = text.rsplit(maxsplit=1)[0]
            nb_words -= 1
        unroll = nb_words // args.unroll_steps

    complete_ivec = ivec_extractor(text)
    complete_ivec_len = length(complete_ivec)

    partial_ivecs = [
        ivec_extractor(" ".join(prefix)) for prefix in [
            text.split()[:l] for l in range(0, nb_words, unroll)
        ]
    ]

    if args.unroll_steps:
        partial_ivecs = torch.stack(partial_ivecs[:args.unroll_steps])

    if partial_ivecs.size(0) != args.unroll_steps:
        print(partial_ivecs.size(0))

    distances = euclidean_distance(partial_ivecs, complete_ivec)
    cos_sims = cosine_similarity(partial_ivecs, complete_ivec)

    return distances, complete_ivec_len, cos_sims


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--unroll', type=int, default=35,
                        help="bptt equivalent. Applies only when --unroll-steps is not set.")
    parser.add_argument('--unroll-steps', type=int, 
                        help="for how many steps the text should unrolled. Overrides --unroll.")
    source_opt = parser.add_mutually_exclusive_group(required=True)
    source_opt.add_argument('--document', help="what to perform the analysis on")
    source_opt.add_argument('--file-list', help="file with list of files analyze")
    parser.add_argument('--ivec-extractor', required=True,
                        help="iVector extractor to use")
    args = parser.parse_args()
    print(args)

    print("loading SMM iVector extractor ...")
    with open(args.ivec_extractor, 'rb') as f:
        ivec_extractor = smm_ivec_extractor.load(f)
    print(ivec_extractor)

    if args.document:
        with open(args.document) as f:
            content = f.read()
        distances = analyze_document(content, ivec_extractor)

        print(distances)

    elif args.file_list:
        if not args.unroll_steps:
            raise ValueError("When analyzing a filelist, --unroll-steps HAS to be specified.")

        documents = filenames_file_to_filenames(args.file_list)

        distances = []
        ci_lens = []
        cos_sims = []
        nb_failed = 0
        for doc in documents:
            with open(doc) as f:
                content = f.read()

            try:
                distance, ci_len, cos_sim = analyze_document(content, ivec_extractor)
                distances.append(distance)
                ci_lens.append(ci_len)
                cos_sims.append(cos_sim)
            except ValueError:
                nb_failed += 1

        if nb_failed > 0:
            sys.stderr.write("Failed analyzing {} documents, because they are too short.\n".format(nb_failed))

        distances = torch.stack(distances)
        ci_lens = torch.stack(ci_lens)
        cos_sims = torch.stack(cos_sims)

        print(torch.stack([distances.min(dim=0)[0], distances.mean(dim=0), distances.max(dim=0)[0], distances.var(dim=0)]).t())

        distances_normed = distances / ci_lens
        print(torch.stack([
            distances_normed.min(dim=0)[0], distances_normed.mean(dim=0),
            distances_normed.max(dim=0)[0], distances_normed.var(dim=0)
        ]).t())

        print(torch.stack([cos_sims.min(dim=0)[0], cos_sims.mean(dim=0), cos_sims.max(dim=0)[0], cos_sims.var(dim=0)]).t())



import torch


class LineTooLongError(Exception):
    pass


def batchify(data, bsz, cuda):
    """ For simple rearranging of 'single sentence' data.
    """
    # Work out how cleanly we can divide the dataset into bsz parts.
    nbatch = data.size(0) // bsz
    # Trim off any extra elements that wouldn't cleanly fit (remainders).
    data = data.narrow(0, 0, nbatch * bsz)
    # Evenly divide the data across the bsz batches.
    data = data.view(bsz, -1).t().contiguous()
    if cuda:
        data = data.cuda()
    return data


class Batcher:
    """Groups sequences in a list into batches
    """
    def __init__(self, samples, max_batch_size=None, max_total_len=None):
        self.max_batch_size = max_batch_size
        self.max_total_len = max_total_len
        self.samples = samples

    def _batch_size_ok(self, i, j):
        if not self.max_batch_size:
            return True

        return (j-i) <= self.max_batch_size

    def _total_len_ok(self, i, j):
        if not self.max_total_len:
            return True

        return max((len(t) for t in self.samples[i:j]), default=0) * (j-i) <= self.max_total_len

    def __iter__(self):
        i = 0
        while i < len(self.samples):
            j = i
            while j <= len(self.samples) and self._batch_size_ok(i, j) and self._total_len_ok(i, j):
                j += 1
            j -= 1

            if i == j:
                raise LineTooLongError(f'Failed to construct a batch on line {i} (zero-based)')

            batch = self.samples[i:j]
            assert len(batch) > 0
            assert sum(len(s) for s in batch) > 0

            yield batch
            i = j


class BatchBuilder():
    def __init__(self, streams, max_batch_size, discard_h=True):
        """ For complex combination of different lenghts sources.
        """
        self._streams = streams

        if max_batch_size <= 0:
            raise ValueError("BatchBuilder must be constructed"
                "with a positive batch size, (got {})".format(max_batch_size)
            )
        self._max_bsz = max_batch_size
        self._discard_h = discard_h

    def __iter__(self):
        streams = [iter(s) for s in self._streams]
        active_streams = []
        reserve_streams = streams

        while True:
            batch = []
            streams_continued = []
            streams_ended = []
            for i, s in enumerate(active_streams):
                try:
                    batch.append(next(s))
                    streams_continued.append(i)
                except StopIteration:
                    streams_ended.append(i)

            active_streams = [active_streams[i] for i in streams_continued]

            # refill the batch (of active streams)
            while len(reserve_streams) > 0:
                if len(batch) == self._max_bsz:
                    break

                stream = reserve_streams[0]
                del reserve_streams[0]
                try:
                    batch.append(next(stream))
                    active_streams.append(stream)
                except StopIteration:
                    pass

            if len(batch) == 0:
                return

            if self._discard_h:
                hs_passed_on = streams_continued
            else:
                hs_passed_on = (streams_continued + streams_ended)[:len(batch)]

            parts = zip(*batch)
            parts = [torch.stack(part) for part in parts]
            yield tuple(parts) + (torch.LongTensor(hs_passed_on), )



#!/usr/bin/env python
import argparse
import math
import sys

import torch


BATCH_SIZE = 1


def per_word_logprobs(words, lm):
    words_tensor = torch.tensor([lm.vocab[w] for w in words], requires_grad=False).view(1, -1)

    x = words_tensor[:, :-1]
    t = words_tensor[:, 1:]

    h0 = lm.model.init_hidden(x.size(0))
    if not lm.model.batch_first:
        x = x.t()
        t = t.t()

    with torch.no_grad():
        out_embs, h = lm.model(x, h0)
        log_probs = -lm.decoder.neg_log_prob_raw(out_embs, t)

    return log_probs.view(-1)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--lm', required=True)
    parser.add_argument('--sb', default='</s>')
    parser.add_argument('--add-boundaries', action='store_true')
    args = parser.parse_args()

    lm = torch.load(args.lm, map_location=lambda storage, location: storage)
    lm.eval()

    nb_sentences = 0
    nb_words = 0
    first_line = True
    total_logprob = 0.0
    sentence_boundary_logprob = 0.0
    for line_no, line in enumerate(sys.stdin):
        words = line.split()

        if len(words) <= 1:
            sys.stderr.write('Skipping line {} due to only containing "{}".\n'.format(line_no, words))
            continue

        if args.add_boundaries:
            if words[-1] == args.sb:
                sys.stderr.write('Skipping line {}, contains sentence boundary, despite --add-boundaries. "{}".'.format(line_no, args.sb))
                continue
            else:
                words.append(args.sb)
        else:
            if words[-1] != args.sb:
                sys.stderr.write('Skipping line {}, lacks sentence boundary "{}".'.format(line_no, args.sb))
                continue

        log_probs = per_word_logprobs([args.sb] + words, lm)

        if first_line:
            first_line = False
        else:
            sys.stdout.write('\n')

        sys.stdout.write(line)
        for w, log_p in zip(words, log_probs):
            word_field = "p( {} | ...)".format(w)
            sys.stdout.write("\t{}\t= [1gram] {:.3f}\n".format(word_field, log_p.item(), log_p.exp().item()))

        nb_sentences += 1
        nb_words += len(words)
        total_logprob += log_probs.sum().item()
        sentence_boundary_logprob += log_probs[-1]

    nb_words -= nb_sentences
    sys.stdout.write('{} sentences, {} words, 0 OOVs\n'.format(nb_sentences, nb_words))
    sys.stdout.write('0 zeroprobs, logprob= {:.3f} ppl= {:.2f} ppl1= {:.2f}\n'.format(
        total_logprob, math.exp(-total_logprob/(nb_words+nb_sentences)),
        math.exp(-(total_logprob-sentence_boundary_logprob)/nb_words)
    ))


if __name__ == '__main__':
    main()



import torch


def categorical_entropy(p, eps=1e-100):
    zeros = p <= eps

    log_p = p.log()
    log_p.masked_fill_(zeros, 0.0) # eliminates -inf for p[x] = 0.0

    H_p = - torch.sum(p*log_p, dim=-1)
    return H_p / torch.log(torch.FloatTensor([2]))
    

def categorical_cross_entropy(p, q, eps=1e-100):
    zeros = p <= eps

    log_q = q.log()
    log_q = torch.zeros_like(p) + log_q
    log_q.masked_fill_(zeros, 0.0) # eliminates -inf for p[x] = 0.0

    Xent = - torch.sum(p*log_q, dim=-1)
    return Xent / torch.log(torch.FloatTensor([2]))


def categorical_kld(p, q):
    return categorical_cross_entropy(p, q) - categorical_entropy(p)



import torch
from .common import TestCase

from brnolm.runtime.runtime_utils import repackage_hidden


class TensorReorganizerTests(TestCase):
    def setUp(self):
        tensor = torch.tensor(
            [[[0.1, 0.1], [0.2, 0.2], [0.3, 0.3]]],
            requires_grad=True
        )
        self.computed_tensor = tensor * 3

    def test_data_kept(self):
        repackaged = repackage_hidden(self.computed_tensor)
        self.assertEqual(self.computed_tensor, repackaged)

    def test_result_requires_grad(self):
        repackaged = repackage_hidden(self.computed_tensor)
        self.assertTrue(repackaged.requires_grad_)

    def test_is_detached(self):
        self.assertFalse(self.computed_tensor.grad_fn is None)
        repackaged = repackage_hidden(self.computed_tensor)
        self.assertTrue(repackaged.grad_fn is None)



from brnolm.runtime.tensor_reorganization import TensorReorganizer

class CheatingIvecAppender():
    def __init__(self, tokens, ivec_eetor):
        """
            Args:
                tokens (TokenizedSplit): Source of tokens, represents single 'document'.
        """
        self.tokens = tokens
        all_words = " ".join(self.tokens.input_words())
        self._ivec = ivec_eetor(all_words)


    def __iter__(self):
        for x, t in self.tokens:
            yield (x, t, self._ivec)


class HistoryIvecAppender():
    def __init__(self, tokens, ivec_eetor):
        """
            Args:
                tokens (TokenizedSplit): Source of tokens, represents single 'document'.
        """
        self.tokens = tokens
        self._ivec_eetor = ivec_eetor


    def __iter__(self):
        history_words = []
        for (x, t), words in zip(self.tokens, self.tokens.input_words()):
            ivec = self._ivec_eetor(" ".join(history_words))
            history_words += words.split()
            yield (x, t, ivec)


class ParalelIvecAppender:
    def __init__(self, stream, extractor, translator):
        self._stream = stream
        self._extractor = extractor
        self._translator = translator 
        self._reorganizer = TensorReorganizer(extractor.zero_bows)

    def __iter__(self):
        old_bows = None
        for x, t, mask in self._stream:
            corresponding_bows = self._reorganizer(old_bows, mask, x.size(0))
            try:
                ivectors = self._extractor(corresponding_bows)
            except RuntimeError:
                print(x.size(), mask, corresponding_bows.size())
                raise
            old_bows = corresponding_bows + self._translator(x)
            yield x, t, ivectors, mask



#!/usr/bin/env python
import argparse
import os

import pickle

import sys
from smm import SMM
import utils

import .smm_ivec_extractor


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--smm', required=True, help="path to a trained SMM model")
    parser.add_argument('--tokenizer', required=True, help="path to the tokenizer used for processing the data for training the SMM")
    parser.add_argument('--save', required=True, help="where to put the final SMM ivec extractor")
    args = parser.parse_args()

    model_f = os.path.realpath(args.smm)
    model, config = utils.load_model_and_config(model_f)

    with open(args.tokenizer, 'rb') as f:
        tokenizer = pickle.load(f)

    ivec_extractor = smm_ivec_extractor.IvecExtractor(model, nb_iters=10, lr=config['eta'], tokenizer=tokenizer)

    with open(args.save, 'wb') as f:
        ivec_extractor.save(f)



from unittest import TestCase

from brnolm.runtime.model_statistics import scaled_int_str


class ScaledIntRepreTests(TestCase):
    def test_order_0(self):
        self.assertEqual(scaled_int_str(0), '0')

    def test_order_1(self):
        self.assertEqual(scaled_int_str(10), '10')

    def test_order_2(self):
        self.assertEqual(scaled_int_str(210), '210')

    def test_order_3(self):
        self.assertEqual(scaled_int_str(3210), '3.2k')

    def test_order_4(self):
        self.assertEqual(scaled_int_str(43210), '43.2k')

    def test_order_5(self):
        self.assertEqual(scaled_int_str(543210), '543.2k')

    def test_order_6(self):
        self.assertEqual(scaled_int_str(6543210), '6.5M')



#!/usr/bin/env python

import argparse
import math
import runtime_utils

import torch

import language_model
import smm_ivec_extractor


def bows_to_ps(bows):
    uni_ps = bows.t() / bows.sum(dim=1)
    return uni_ps.t()

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--file-list')
    parser.add_argument('--ivec-lm')
    parser.add_argument('--ivec-extractor')
    args = parser.parse_args()

    print("loading LM...")
    with open(args.ivec_lm, 'rb') as f:
        lm = language_model.load(f)
    lm.model.cuda()
    print(lm.model)

    print("loading SMM iVector extractor ...")
    with open(args.ivec_extractor, 'rb') as f:
        ivec_extractor = smm_ivec_extractor.load(f)
    print(ivec_extractor)

    fns = runtime_utils.filenames_file_to_filenames(args.file_list)
    documents = []
    for fn in fns:
        with open(fn) as f:
            documents.append(f.read().split())

    bows = torch.zeros(len(documents), len(lm.vocab)).long()
    for doc_no, doc in enumerate(documents):
        for w in doc:
            bows[doc_no, lm.vocab[w]] += 1
    unigram_ps = bows_to_ps(bows.float()).cuda()

    cross_entropies = []
    for doc_no, doc in enumerate(documents):
        text = " ".join(doc)
        ivec = ivec_extractor(text).cuda()
        qs = lm.model.ivec_to_logprobs(ivec).data
        cross_entropies.append(unigram_ps[doc_no] @ qs)

    cross_entropies = torch.FloatTensor(cross_entropies)
    avg_ce = -cross_entropies @ bows.float().sum(dim=1) / bows.sum()

    print("{:.4f} {:.2f}".format(avg_ce, math.exp(avg_ce)))



from unittest import TestCase

from brnolm.oov_clustering.det import det_points_from_score_tg
from brnolm.oov_clustering.det import subsample_list
from brnolm.oov_clustering.det import area_under_curve, eer


class DetPointTests(TestCase):
    def test_trivial(self):
        score_tg = [
            (0.0, 0),
            (1.0, 1)
        ]

        det_points = [
            [0.5, 0.0],
            [0.0, 0.0],
            [0.0, 0.5],
        ]

        self.assertEqual(det_points_from_score_tg(score_tg)[0], det_points)

    def test_permutation(self):
        score_tg = [
            (1.0, 1),
            (0.0, 0),
        ]

        det_points = [
            [0.5, 0.0],
            [0.0, 0.0],
            [0.0, 0.5],
        ]

        self.assertEqual(det_points_from_score_tg(score_tg)[0], det_points)

    def test_simple_bad_system(self):
        score_tg = [
            (1.0, 0),
            (0.0, 1),
        ]

        det_points = [
            [0.5, 0.0],
            [0.5, 0.5],
            [0.0, 0.5],
        ]

        self.assertEqual(det_points_from_score_tg(score_tg)[0], det_points)

    def test_multiple_points(self):
        score_tg = [
            (0.0, 0),
            (0.1, 1),
            (0.2, 0),
            (0.3, 1),
            (0.4, 1),
        ]

        det_points = [
            [0.6, 0.0],
            [0.4, 0.0],
            [0.2, 0.0],
            [0.2, 0.2],
            [0.0, 0.2],
            [0.0, 0.4],
        ]

        self.assertEqual(det_points_from_score_tg(score_tg)[0], det_points)


class ListSubsamplingTests(TestCase):
    def test_trivial(self):
        self.assertEqual(
            subsample_list([1, 2], 2),
            [1, 2]
        )

    def test_halving_longer(self):
        self.assertEqual(
            subsample_list([1, 2, 3, 4], 2),
            [1, 4]
        )

    def test_halving_uneven(self):
        self.assertEqual(
            subsample_list([1, 2, 3, 4, 5], 2),
            [1, 5]
        )

    def test_halving_uneven_2(self):
        self.assertEqual(
            subsample_list([1, 2, 3, 4, 5], 3),
            [1, 3, 5]
        )

    def test_every_third(self):
        self.assertEqual(
            subsample_list([1, 2, 3, 4, 5, 6], 2),
            [1, 6]
        )

    def test_include_most(self):
        self.assertIn(
            subsample_list([1, 2, 3, 4, 5, 6], 5),
            [
                [1, 3, 4, 5, 6],
                [1, 2, 4, 5, 6],
                [1, 2, 3, 5, 6],
                [1, 2, 3, 4, 6],
            ]
        )

    def test_include_4_of_6(self):
        self.assertIn(
            subsample_list([1, 2, 3, 4, 5, 6], 4),
            [
                [1] + middle + [6] for middle in [
                    [2, 3], [2, 4], [2, 5],
                    [3, 4], [3, 5],
                    [4, 5],
                ]
            ]
        )


class AreaComputationTests(TestCase):
    def test_trivial(self):
        self.assertAlmostEqual(area_under_curve([0.0, 1.0], [1.0, 0.0]), 0.5)

    def test_breakpoint(self):
        self.assertAlmostEqual(area_under_curve([0.0, 0.1, 1.0], [1.0, 0.1, 0.0]), 0.1*0.1 + 2*0.1*0.9/2)

    def test_end_addition(self):
        self.assertAlmostEqual(area_under_curve([0.1], [0.1]), 0.1*0.1 + 2*0.1*0.9/2)


class EerComputationTests(TestCase):
    def test_trivial(self):
        self.assertAlmostEqual(eer([0.0, 1.0], [1.0, 0.0]), 0.5)

    def test_skewed(self):
        self.assertAlmostEqual(eer([0.0, 1.0], [0.5, 0.0]), 1.0/3.0)

    def test_exact_hit(self):
        self.assertAlmostEqual(eer([0.0, 0.1, 1.0], [1.0, 0.1, 0.0]), 0.1)



import torch


def word_splitter(line):
    return line.split()


def char_splitter(line, sentence_end_token=None):
    chars = list(line)

    if sentence_end_token is None:
        return chars
    else:
        return chars + [sentence_end_token]


def tokens_from_file(f, vocab, randomize, regime='words'):
    ids = []

    lines = f.read().split('\n')

    if regime == 'words':
        tokenizer = word_splitter
    elif regime == 'chars':
        tokenizer = lambda line: char_splitter(line, '</s>')
    else:
        raise ValueError("unsupported regime {}".format(regime))

    if randomize:
        import random
        random.shuffle(lines)

    for line in lines:
        tokens = tokenizer(line)
        ids.extend([vocab[e] for e in tokens])

    return torch.LongTensor(ids)


def tokens_from_fn(fn, vocab, randomize, regime='words'):
    with open(fn, 'r') as f:
        return tokens_from_file(f, vocab, randomize, regime)


def get_independent_lines(f, vocab):
    lines = []
    for line in f:
        words = line.split()
        if words:
            lines.append(torch.tensor([vocab[w] for w in words]))

    return lines



import unittest
from .common import TestCase
import os
import sys

try:
    import brnolm.smm_itf.smm_ivec_extractor as smm_ivec_extractor
except ImportError:
    sys.stderr.write('Failed to import SMM implementation\n')
from brnolm.language_models.vocab import Vocabulary

import torch
import torch.nn as nn
from torch.autograd import Variable
import numpy as np

from sklearn.feature_extraction.text import CountVectorizer


class DummySMM(nn.Module):
    def __init__(self, ivec_dim):
        self.W = Variable(torch.zeros(ivec_dim, 1), requires_grad=True)
        self.T = Variable(torch.zeros(ivec_dim, 20), requires_grad=True)


@unittest.skipIf(os.environ.get('TEST_SMM') != 'yes', "For SMM tests, set TEST_SMM='yes'")
class IvecExtractorTests(TestCase):
    def setUp(self):
        self.documents_six = ["text consisting of SIX different words", "text"]

        # note that following also contains puctuation marks, to be ignored by CountVectorizer
        self.documents_seven = ["text consisting of SEVEN different words", ", seventh ."]

    def build_neededs(self, documents):
        smm = DummySMM(ivec_dim=4)
        cvect = CountVectorizer(documents, strip_accents='ascii', analyzer='word')
        cvect.fit(documents)
        vocab = cvect.get_feature_names()
        self.cvect = CountVectorizer(documents, strip_accents='ascii', analyzer='word', vocabulary=vocab)

        self.extractor = smm_ivec_extractor.IvecExtractor(smm, nb_iters=10, lr=0.1, tokenizer=self.cvect)
        self.vocab = Vocabulary(unk_word="<unk>", unk_index=0)
        self.vocab.add_from_text(" ".join(documents))

    def test_one_empty_bow(self):
        self.build_neededs(self.documents_six)

        ivecs = self.extractor.zero_bows(1)
        expectation = torch.zeros(1, 6)
        self.assertEqual(ivecs, expectation)

    def test_two_empty_bows(self):
        self.build_neededs(self.documents_six)
        ivecs = self.extractor.zero_bows(2)
        expectation = torch.zeros(2, 6)
        self.assertEqual(ivecs, expectation)

    def test_build_translator_single_word_translation(self):
        self.build_neededs(self.documents_six)
        translator = self.extractor.build_translator(self.vocab)
        word = "SIX"
        lm_word = torch.LongTensor([self.vocab[word]])
        cv_word = torch.from_numpy(self.cvect.transform([word]).A.astype(np.float32)).squeeze()
        self.assertEqual(translator(lm_word), cv_word)

    def test_build_translator_two_word_translation(self):
        self.build_neededs(self.documents_six)
        translator = self.extractor.build_translator(self.vocab)
        words = "SIX words"
        lm_words = torch.LongTensor([self.vocab[w] for w in words.split()])
        cv_words = torch.from_numpy(self.cvect.transform([words]).A.astype(np.float32)).squeeze()
        self.assertEqual(translator(lm_words), cv_words)

    def test_build_translator_two_two_word_translations(self):
        self.build_neededs(self.documents_six)
        translator = self.extractor.build_translator(self.vocab)
        words = ["SIX words", "of text"]
        lm_words = torch.LongTensor([[self.vocab[w] for w in seq.split()] for seq in words])
        cv_words = torch.from_numpy(self.cvect.transform(words).A.astype(np.float32)).squeeze()
        self.assertEqual(translator(lm_words), cv_words)

    def test_build_translator_two_two_word_translations_different_vocab(self):
        self.build_neededs(self.documents_seven)
        translator = self.extractor.build_translator(self.vocab)
        words = ["SIX words", "of text"]
        lm_words = torch.LongTensor([[self.vocab[w] for w in seq.split()] for seq in words])
        cv_words = torch.from_numpy(self.cvect.transform(words).A.astype(np.float32)).squeeze()
        self.assertEqual(translator(lm_words), cv_words)



import pathlib
import torch
import math


class ValidationWatcher:
    def __init__(self, val_fn, initial_val_loss, freq_in_tokens, workdir, lm):
        self.val_losses = [initial_val_loss]
        self.validation_fn = val_fn
        self.lm = lm

        pathlib.Path(workdir).mkdir(parents=True, exist_ok=True)

        assert(freq_in_tokens > 0)
        self.freq = freq_in_tokens

        self.report_fn = '{}/validation-report.txt'.format(workdir)
        self.best_model_fn = '{}/best.lm'.format(workdir)
        self.latest_model_fn = '{}/latest.lm'.format(workdir)

        self.running_loss = 0.0
        self.running_targets = 0
        self.running_updates = 0

        self.nb_total_updates = 0

    def log_training_update(self, loss, nb_targets):
        self.running_loss += loss
        self.running_targets += nb_targets
        self.running_updates += 1
        self.nb_total_updates += 1

        if self.running_targets > self.freq:
            val_loss = self.run_validation()

            running_ppl = math.exp(self.running_loss / self.running_updates)
            val_ppl = math.exp(val_loss)

            desc = '{} updates: {:.2f} {:.2f} {:.3f}\n'.format(
                self.nb_total_updates, running_ppl, val_ppl, val_ppl - running_ppl
            )
            with open(self.report_fn, 'a') as f:
                f.write(desc)

            torch.save(self.lm, self.latest_model_fn)
            if min(self.val_losses) == self.val_losses[-1]:
                torch.save(self.lm, self.best_model_fn)

            self.running_loss = 0.0
            self.running_targets = 0
            self.running_updates = 0

    def run_validation(self):
        val_loss = self.validation_fn()
        self.val_losses.append(val_loss)
        return val_loss



#!/usr/bin/env python

import torch
import argparse
import plotting
import numpy as np


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("model")

    args = parser.parse_args()

    with open(args.model, 'rb') as f:
        model = torch.load(f)

    cs_numpied = [c.weight.data.numpy() for c in model._cs]
    fig_titles = []
    for c in cs_numpied:
        norms = np.linalg.norm(c, ord=2, axis=1)
        fig_titles.append(", ".join([str(x) for x in [np.min(norms), np.mean(norms), np.max(norms)]]))
    plotting.grid_plot(cs_numpied, lambda x: x, "C-Weights", fig_titles)

    plotting.grid_plot(model._cs, lambda c: c.bias.data.numpy()[...,None], "Biases")

    ps_numpied = [p.weight.data.numpy() for p in model._ps]
    fig_titles = []
    for p in ps_numpied:
        norms = np.linalg.norm(p, ord=2, axis=1)
        fig_titles.append(", ".join([str(x) for x in [np.min(norms), np.mean(norms), np.max(norms)]]))
    plotting.grid_plot(ps_numpied, lambda x: x, "P-Weights", fig_titles)



from collections import Mapping
import re


class IndexGenerator():
    def __init__(self, assigned):
        self.next_ = 0
        self.assigned_ = assigned

    def next(self):
        while self.next_ in self.assigned_:
            self.next_ += 1

        retval = self.next_
        self.next_ += 1
        return retval


class Vocabulary(Mapping):
    def __init__(self, unk_word, unk_index):
        self.w2i_ = {unk_word: unk_index}
        self.i2w_ = {unk_index: unk_word}
        self.unk_index_ = unk_index
        self.unk_word_ = unk_word
        self.ind_gen_ = IndexGenerator([unk_index])

    def add_from_text(self, text):
        assert self.ind_gen_

        words = text.split()
        for word in words:
            self.add_word(word)

    def add_word(self, word):
        if word not in self.w2i_:
            index = self.ind_gen_.next()
            self.w2i_[word] = index
            self.i2w_[index] = word
        else:
            pass  # do not do anything for known words

    def is_continuous(self):
        return max(self.i2w_.keys()) == len(self) - 1

    def missing_indexes(self):
        return [i for i in range(len(self)) if i not in self.i2w_]

    def w2i(self, word):
        return self.w2i_.get(word, self.unk_index_)

    @property
    def unk_word(self):
        return self.unk_word_

    @property
    def unk_ind(self):
        return self.unk_index_

    def __getitem__(self, idx):
        return self.w2i(idx)

    def i2w(self, index):
        return self.i2w_[index]

    def __len__(self):
        return len(self.w2i_)

    def __iter__(self):
        return iter(self.w2i_)


def vocab_from_kaldi_wordlist_base(f, unk_word, word_re, remove_quotes):
    d = {}
    line_re = re.compile('\s*(?P<word>' + word_re + ')\s+(?P<ind>[0-9]+)\s*\n?')
    for i, line in enumerate(f):
        m = line_re.fullmatch(line)

        if m is None:
            raise ValueError("Weird line {}: '{}'".format(i, line))

        w = m.group('word')
        if remove_quotes:
            assert w[0] == w[-1] == "'"
            w = w[1:-1]
        i = int(m.group('ind'))
        assert i >= 0

        if w in d:
            raise ValueError(f'Attempt to redefince "{w}" to {i}, while it already has {d[w]} assigned')
        d[w] = i

    try:
        vocab = Vocabulary(unk_word, d[unk_word])
    except KeyError:
        raise ValueError("Unk word {} not present in the kaldi wordlist!".format(unk_word))
    vocab.ind_gen_ = None
    vocab.w2i_ = d
    for w in d:
        vocab.i2w_[d[w]] = w

    vocab.size_ = max(d.values()) + 1

    return vocab


def vocab_from_kaldi_wordlist(f, unk_word='<unk>'):
    return vocab_from_kaldi_wordlist_base(f, unk_word, word_re='\S+', remove_quotes=False)


def quoted_vocab_from_kaldi_wordlist(f, unk_word='<unk>'):
    return vocab_from_kaldi_wordlist_base(f, unk_word, word_re="'.+'", remove_quotes=True)



import torch

from brnolm.data_pipeline.temporal_splitting import TemporalSplits


class TokenizedSplitFFBase():
    def __init__(self, f, vocab, temporal_split_builder):
        """
            Args:
                f (file): File with a document.
                vocab (Vocabulary): Vocabulary for translation word -> index
        """
        sentence = f.read()
        self._words = sentence.split()
        self._tokens = torch.LongTensor([vocab[w] for w in self._words])

        self._temp_splits = temporal_split_builder(self._tokens)

    def __iter__(self):
        for x, t in self._temp_splits:
            yield x, t

    def __len__(self):
        return len(self._temp_splits)

    def input_words(self):
        for lend, rend in self._temp_splits.ranges():
            yield " ".join(self._words[lend:rend])


class TokenizedSplit(TokenizedSplitFFBase):
    def __init__(self, f, vocab, unroll_length):
        """
            Args:
                f (file): File with a document.
                vocab (Vocabulary): Vocabulary for translation word -> index
        """
        ts_builder = lambda seq: TemporalSplits(seq, nb_inputs_necessary=1, nb_targets_parallel=unroll_length)
        super().__init__(f, vocab, ts_builder)


class TokenizedSplitSingleTarget(TokenizedSplitFFBase):
    def __init__(self, f, vocab, unroll_length):
        """
            Args:
                f (file): File with a document.
                vocab (Vocabulary): Vocabulary for translation word -> index
        """
        ts_builder = lambda seq: TemporalSplits(seq, nb_inputs_necessary=unroll_length, nb_targets_parallel=1)
        super().__init__(f, vocab, ts_builder)


class TokenizedSplitFFMultiTarget(TokenizedSplitFFBase):
    def __init__(self, f, vocab, hist_len, nb_targets_parallel):
        """
            Args:
                f (file): File with a document.
                vocab (Vocabulary): Vocabulary for translation word -> index
        """
        ts_builder = lambda seq: TemporalSplits(seq, nb_inputs_necessary=hist_len, nb_targets_parallel=nb_targets_parallel)
        super().__init__(f, vocab, ts_builder)


class DomainAdaptationSplitFFBase:
    def __init__(self, f, vocab, end_portion, ts_builder):
        sentence = f.read()
        words = sentence.split()

        nb_domain_words = int(len(words)*end_portion-0.01)

        self._tokens = torch.LongTensor([vocab[w] for w in words[:-nb_domain_words]])
        self._domain_string = " ".join(words[len(words)-nb_domain_words:])

        self._temp_splitter = ts_builder(self._tokens)

    def __iter__(self):
        for x, t in self._temp_splitter:
            yield x, t

    def __len__(self):
        return len(self._temp_splitter)

    def input_words(self):
        return [self._domain_string]


class DomainAdaptationSplitFFMultiTarget(DomainAdaptationSplitFFBase):
    def __init__(self, f, vocab, hist_len, nb_targets_parallel, end_portion):
        """
            Args:
                f (file): File with a document.
                vocab (Vocabulary): Vocabulary for translation word -> index
        """
        ts_builder = lambda seq: TemporalSplits(seq, nb_inputs_necessary=hist_len, nb_targets_parallel=nb_targets_parallel)
        super().__init__(f, vocab, end_portion, ts_builder)


class DomainAdaptationSplit(DomainAdaptationSplitFFBase):
    def __init__(self, f, vocab, unroll_length, end_portion):
        """
            Args:
                f (file): File with a document.
                vocab (Vocabulary): Vocabulary for translation word -> index
        """

        ts_builder = lambda seq: TemporalSplits(seq, nb_inputs_necessary=unroll_length, nb_targets_parallel=1)
        super().__init__(f, vocab, end_portion, ts_builder)



import torch

from typing import Dict, Any


class Singleton(type):
    _instances: Dict[Any, Any] = {}  # TODO what is the actual type?

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class InfiniNoneType(metaclass=Singleton):
    def __eq__(self, other):
        return other is None or isinstance(other, InfiniNone)

    def __iter__(self):
        while True:
            yield InfiniNoneType()


InfiniNone = InfiniNoneType()


def reorg_single(orig, mask, new=None):
    reorg = torch.index_select(orig, dim=-2, index=mask)
    if new is not InfiniNone:
        reorg = torch.cat([reorg, new], dim=-2)

    return reorg


class TensorReorganizer():
    def __init__(self, zeros_provider):
        self._zeros_provider = zeros_provider

    def __call__(self, orig, mask, batch_size):
        if len(mask) == 0:
            return self._zeros_provider(batch_size)

        if mask.size(0) > batch_size:
            raise ValueError("Cannot reorganize mask {} to batch size {}".format(mask, batch_size))

        if isinstance(orig, tuple):
            single_var = False
        elif isinstance(orig, torch.Tensor):
            single_var = True
        else:
            raise TypeError(
                "orig has unsupported type {}, "
                "only tuples and Tensors are accepted".format(
                    orig.__class__
                )
            )

        adding = mask.size(0) < batch_size
        if adding:
            nb_needed_new = batch_size - mask.size(0)
            new = self._zeros_provider(nb_needed_new)
        else:
            new = InfiniNone

        if single_var:
            reorg = reorg_single(orig, mask, new)
        else:
            reorg = tuple(reorg_single(o, mask, n) for o, n in zip(orig, new))

        return reorg



import matplotlib.pyplot as plt
import atexit

atexit.register(plt.show)

DEFAULT_COLORING = {'cmap':plt.cm.RdBu, 'vmin':-0.5, 'vmax':0.5, 'color_bar': True}

def _flip_ord(X, xsize, ysize):
    assert len(X) == xsize * ysize
    reord_X = []
    for i in range(len(X)):
        x_y = i // xsize
        x_x = i % xsize
        reord_X.append(X[x_x * ysize + x_y])

    return reord_X
    

def grid_plot(X, numpy_accessor, title, fig_titles=None, 
              coloring=DEFAULT_COLORING):

    if fig_titles == None:
        fig_titles = [""] * len(X)
    assert(len(X) == len(fig_titles))

    reord_X = _flip_ord(X, 5, 3)

    fig, axes = plt.subplots(nrows=3, ncols=5, figsize=(20,20))
    for x, ax, f_title in zip(X, axes.flat, fig_titles):
        im = ax.imshow(
            numpy_accessor(x), 
            cmap=coloring['cmap'], 
            vmin=coloring['vmin'], 
            vmax=coloring['vmax'],
        )
        ax.set_title(f_title, fontsize=8)
        ax.axis('off')
    if coloring['colorbar']:
        fig.colorbar(im, ax=axes.ravel().tolist())
    fig.canvas.set_window_title(title)
    plt.show(block=False)



#!/usr/bin/env python
import argparse
import torch

import smm_ivec_extractor

from runtime_utils import init_seeds, filenames_file_to_filenames


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='PyTorch RNN/LSTM Language Model')
    parser.add_argument('--filelist', type=str, required=True,
                        help='file with paths to documents')
    parser.add_argument('--seed', type=int, default=1111,
                        help='random seed')
    parser.add_argument('--cuda', action='store_true',
                        help='use CUDA')
    parser.add_argument('--ivec-extractor', type=str, required=True,
                        help='where to load a ivector extractor from')
    args = parser.parse_args()
    print(args)

    init_seeds(args.seed, args.cuda)

    print("loading SMM iVector extractor ...")
    with open(args.ivec_extractor, 'rb') as f:
        ivec_extractor = smm_ivec_extractor.load(f)
    print(ivec_extractor)

    print("reading data...")
    filenames = filenames_file_to_filenames(args.filelist)
    texts = []
    for fn in filenames:
        with open(fn) as f:
            texts.append(f.read())

    print("computing iVectors...")
    ivecs = [ivec_extractor(t) for t in texts]
    ivecs = torch.stack(ivecs)

    print("Elements mean:\t", ivecs.mean())
    print("Elements var:\t", ivecs.var())
    
    sq_magnitudes = ivecs.pow(2).sum(1)
    print("Sq magn mean:\t", sq_magnitudes.mean())
    print("Sq magn var:\t", sq_magnitudes.var())



import torch
import torch.nn as nn
from typing import Tuple


class LSTMLanguageModel(nn.Module):
    """Container module with an encoder, a recurrent module, and a decoder."""

    def __init__(self, token_encoder, dim_input, dim_lstm, nb_layers, dropout=0.5):
        super(LSTMLanguageModel, self).__init__()
        self.drop = nn.Dropout(dropout)
        self.encoder = token_encoder
        self.rnn = nn.LSTM(dim_input, dim_lstm, nb_layers, dropout=dropout, batch_first=True)

        self.dim_lstm = dim_lstm
        self.nlayers = nb_layers
        self.nhid = dim_lstm

        self.batch_first = True
        self.in_len = 1

    def init_weights(self):
        initrange = 0.1
        self.encoder.weight.data.uniform_(-initrange, initrange)

    def forward(self, input, hidden: Tuple[torch.Tensor, torch.Tensor]):
        emb = self.drop(self.encoder(input))
        output, hidden = self.rnn(emb, hidden)
        output = self.drop(output)
        return output, hidden

    def output_expected_embs(self, input):
        assert (len(input.size()) == 2)  # batch X time index
        assert (input.size()[0] == 1)

        hidden = self.init_hidden(1)
        emb = self.drop(self.encoder(input))
        outputs, _ = self.rnn(emb, hidden)
        return outputs

    @torch.jit.export
    def init_hidden(self, bsz: int):
        weight = self.state_dict()['rnn.weight_ih_l0']
        return (weight.new_zeros(self.nlayers, bsz, self.nhid),
                weight.new_zeros(self.nlayers, bsz, self.nhid))

    @torch.jit.export
    def extract_output_from_h(self, hidden_state: Tuple[torch.Tensor, torch.Tensor]):
        h = hidden_state[0]  # hidden state is (h, c)
        return h[-1]  # last layer is the output one


class LSTMPLanguageModel(nn.Module):
    """Container module with an encoder, a recurrent module, and a decoder."""

    def __init__(self, ntoken, ninp, nhid, nlayers, dropout=0.5, tie_weights=False):
        super().__init__()
        self.drop = nn.Dropout(dropout)
        self.encoder = nn.Embedding(ntoken, ninp)
        self.rnn = nn.LSTM(ninp, nhid, nlayers, dropout=dropout, batch_first=True)
        self.proj = nn.Linear(nhid, ninp)

        if tie_weights:
            raise NotImplementedError

        self.init_weights()

        self.nhid = nhid
        self.nlayers = nlayers

        self.batch_first = True
        self.in_len = 1

    def init_weights(self):
        initrange = 0.1
        self.encoder.weight.data.uniform_(-initrange, initrange)

    def forward(self, input, hidden):
        emb = self.drop(self.encoder(input))
        output, hidden = self.rnn(emb, hidden)
        output = self.drop(output)
        projected = self.proj(output)
        return projected, hidden

    def init_hidden(self, bsz):
        weight = next(self.parameters()).data
        return (weight.new_zeros(self.nlayers, bsz, self.nhid),
                weight.new_zeros(self.nlayers, bsz, self.nhid))



import numpy as np
import copy


def area_under_curve(xs_in, ys_in):
    assert(len(xs_in) == len(ys_in))

    xs = list(copy.deepcopy(xs_in))
    ys = list(copy.deepcopy(ys_in))

    if xs[0] > 0.0:
        xs.insert(0, 0.0)
        ys.insert(0, 1.0)

    if ys[-1] > 0.0:
        xs.append(1.0)
        ys.append(0.0)

    running_sum = 0.0

    for i in range(len(xs)-1):
        x_len = xs[i+1] - xs[i]
        avg_y = (ys[i] + ys[i+1])/2
        running_sum += x_len * avg_y

    return running_sum


def eer(xs, ys):
    assert(len(xs) == len(ys))

    eer = float('nan')
    for i in range(len(xs)-1):
        if xs[i] < ys[i] and xs[i+1] >= ys[i+1]:
            d_i = abs(xs[i] - ys[i])
            d_ip1 = xs[i+1] - ys[i+1]
            lambda_i = d_i/(d_i + d_ip1)
            eer = (1.0 - lambda_i) * xs[i] + lambda_i * xs[i+1]

    return eer


def det_points_from_score_tg(score_tg):
    nb_trials = len(score_tg)
    nb_same = sum(s[1] for s in score_tg)
    nb_different = nb_trials - nb_same

    sorted_score_tg = sorted(score_tg, key=lambda s: s[0], reverse=True)

    nb_correct_same = 0
    nb_correct_different = nb_different
    nb_false_alarms = 0
    nb_misses = nb_same
    mis_fas = [[nb_misses/nb_trials, nb_false_alarms/nb_trials]]

    for s in sorted_score_tg:
        if s[1] == 1:
            nb_misses -= 1
            nb_correct_same += 1
        else:
            nb_false_alarms += 1
            nb_correct_different -= 1

        mis_fas.append([nb_misses/nb_trials, nb_false_alarms/nb_trials])

    return mis_fas, [s[0] for s in sorted_score_tg]


def subsampling_indices(length, max_points):
    ''' Ensures that both the first and the last element are included.
    '''
    all_indices = list(range(length))

    subsampling_coeff_exact = (len(all_indices) - 1) / (max_points-1)
    if subsampling_coeff_exact > 1.0 and subsampling_coeff_exact < 2.0:
        inverse_subsampling_coeff_exact = (len(all_indices) - 2) / (len(all_indices) - max_points)
        inverse_subsampling_coeff = int(inverse_subsampling_coeff_exact)
        to_drop = list(range(1, len(all_indices)-1, inverse_subsampling_coeff))
        return [x for i, x in enumerate(all_indices) if i not in to_drop]
    else:
        subsampling_coeff = int(subsampling_coeff_exact)
        return all_indices[0:-1:subsampling_coeff] + [all_indices[-1]]


def pick(the_list, indices):
    return [the_list[i] for i in indices]


def subsample_list(the_list, max_points):
    ''' Ensures that both the first and the last element are included.
    '''
    indices = subsampling_indices(len(the_list), max_points)
    return pick(the_list, indices)


class DETCurve:
    def __init__(self, score_tg, baseline, max_det_points=0):
        self._baseline = baseline

        nb_trials = len(score_tg)
        nb_same = sum(s[1] for s in score_tg)
        nb_different = nb_trials - nb_same

        self._max_miss_rate = nb_same / nb_trials
        self._max_fa_rate = nb_different / nb_trials

        print("# positive trials: {} ({:.1f} %)".format(nb_same, 100.0*nb_same/nb_trials))
        print("# negative trials: {} ({:.1f} %)".format(nb_different, 100.0*nb_different/nb_trials))

        mis_fas, sorted_scores = det_points_from_score_tg(score_tg)

        if max_det_points > 0 and max_det_points < len(mis_fas):
            self._nb_clusterings = subsampling_indices(len(mis_fas), max_det_points)
            self._scores = pick(sorted_scores, self._nb_clusterings[:-1] + [len(sorted_scores)-1])
            mis_fas = pick(mis_fas, self._nb_clusterings)
        else:
            self._nb_clusterings = range(len(mis_fas))
            self._scores = sorted_scores

        self._miss_rate = [msfa[0] for msfa in mis_fas]
        self._fa_rate = [msfa[1] for msfa in mis_fas]

    def textual_report(self):
        report = ""
        area_line_fmt = "Area under DET curve (in linspace): {:.5f}"
        eer_line_fmt = "EER: {:.2f} %"

        system_au_det = area_under_curve(self._fa_rate, self._miss_rate)
        system_eer = eer(self._fa_rate, self._miss_rate)

        if self._baseline:
            area_line_fmt += " / {:.5f} / {:.2f} %"
            eer_line_fmt += " / {:.2f} % / {:.2f} %"

            baseline_au_det = self._max_miss_rate * self._max_fa_rate / 2.0
            baseline_eer = self._max_miss_rate * self._max_fa_rate / (self._max_miss_rate + self._max_fa_rate)

            report += area_line_fmt.format(
                system_au_det,
                baseline_au_det,
                100.0 * (1.0 - system_au_det/baseline_au_det)
            ) + '\n'
            report += eer_line_fmt.format(
                100.0*system_eer,
                100.0*baseline_eer,
                100.0 * (1.0 - system_eer/baseline_eer)
            ) + '\n'

        else:
            report += area_line_fmt.format(system_au_det) + '\n'
            report += eer_line_fmt.format(100.0*system_eer) + '\n'

        return report

    def plot(self, log_axis, scaled_axis, eer_line, filename):
        import matplotlib.pyplot as plt
        fig = plt.figure()
        ax = fig.add_subplot(111)

        if log_axis:
            plt_func = plt.loglog
        else:
            plt_func = plt.plot
        plt_func(self._miss_rate, self._fa_rate, label='System')

        for i, score, xy in zip(self._nb_clusterings, self._scores, zip(self._miss_rate, self._fa_rate)):
            label = '{} ({:.2f})'.format(i, score)
            ax.annotate((label), xy=xy, textcoords='data', fontsize='small')

        if self._baseline:
            xs = np.linspace(0, self._max_miss_rate)
            ys = np.linspace(self._max_fa_rate, 0)
            plt_func(xs, ys, label='Baseline')

            plt.legend()

        if eer_line:
            endpoint = min([self._max_fa_rate, self._max_miss_rate])
            plt_func([0.0, endpoint], [0.0, endpoint], color='k', linestyle='-.', linewidth=0.75)

        if scaled_axis:
            plt.axis('scaled')
        plt.xlim(left=0.0)
        plt.ylim(bottom=0.0)
        plt.xlabel('miss rate')
        plt.ylabel('FA rate')

        if filename:
            plt.savefig(filename)
        else:
            plt.show()



import pickle
import yaml

from brnolm.data_pipeline.reading import tokens_from_fn
# from brnolm.data_pipeline.multistream import batchify
# from brnolm.data_pipeline.temporal_splitting import TemporalSplits
from brnolm.data_pipeline.threaded import OndemandDataProvider

from brnolm.data_pipeline.aug_paper_pipeline import CleanStreamsProvider, LazyBatcher, TemplSplitterClean
from brnolm.data_pipeline.aug_paper_pipeline import Corruptor
from brnolm.data_pipeline.aug_paper_pipeline import StatisticsCorruptor, Confuser
from brnolm.data_pipeline.aug_paper_pipeline import TargetCorruptor
from brnolm.data_pipeline.aug_paper_pipeline import InputTargetCorruptor

from brnolm.runtime.runtime_utils import TransposeWrapper


def yaml_factory(yaml_fn, lm, device):
    with open(yaml_fn) as f:
        config = yaml.load(f, Loader=yaml.CLoader)

    corruptor_config = config.get('corruptor', None)

    return plain_factory(
        data_fn=config['file'],
        lm=lm,
        tokenize_regime=config['tokenize_regime'],
        batch_size=config['batch_size'],
        device=device,
        target_seq_len=config['target_seq_len'],
        corruptor_config=corruptor_config,
    )


def plain_factory(data_fn, lm, tokenize_regime, batch_size, device, target_seq_len, corruptor_config=None):
    train_ids = tokens_from_fn(data_fn, lm.vocab, randomize=False, regime=tokenize_regime)
    nb_batches = len(train_ids) // batch_size
    train_streams_provider = CleanStreamsProvider(train_ids)

    if corruptor_config:
        train_streams_provider = corruptor_factory(corruptor_config, lm, train_streams_provider)

    batch_former = LazyBatcher(batch_size, train_streams_provider)
    if lm.model.in_len == 1:
        train_data = TemplSplitterClean(target_seq_len, batch_former)
    else:
        raise NotImplementedError("Current data pipeline only supports `in_len==1`.")
    train_data = TransposeWrapper(train_data)
    return OndemandDataProvider(train_data, device), nb_batches


def corruptor_factory(config, lm, input_streams_provider):
    if config['type'] == 'input-0gram':
        subs_rate = float(config['substitution-rate'])
        del_rate = float(config['deletion-rate'])
        ins_rate = float(config['insertion-rate'])

        corruptor = Corruptor(
            input_streams_provider,
            subs_rate,
            len(lm.vocab),
            del_rate,
            ins_rate,
            protected=[lm.vocab['</s>']]
        )
        return corruptor

    elif config['type'] == 'input-1gram':
        stats_filename = config['statistics']
        ins_rate = float(config['insertion-rate'])
        mincount = int(config['mincount'])

        with open(stats_filename, 'rb') as f:
            summary = pickle.load(f)
        confuser = Confuser(summary.confusions, lm.vocab, mincount=mincount)

        corrupted_provider = StatisticsCorruptor(
            input_streams_provider,
            confuser,
            ins_rate,
            protected=[lm.vocab['</s>']],
        )

        return corrupted_provider

    if config['type'] == 'target-0gram':
        subs_rate = float(config['substitution-rate'])
        del_rate = float(config['deletion-rate'])
        ins_rate = float(config['insertion-rate'])

        corrupted_provider = TargetCorruptor(
            input_streams_provider,
            subs_rate,
            len(lm.vocab),
            del_rate,
            ins_rate,
            protected=[lm.vocab['</s>']]
        )
        return corrupted_provider

    if config['type'] == 'input_target-0gram':
        in_subs_rate = float(config['input-substitution-rate'])
        target_subs_rate = float(config['target-substitution-rate'])
        del_rate = float(config['deletion-rate'])
        ins_rate = float(config['insertion-rate'])

        corrupted_provider = InputTargetCorruptor(
            input_streams_provider,
            in_subs_rate,
            target_subs_rate,
            len(lm.vocab),
            del_rate,
            ins_rate,
            protected=[lm.vocab['</s>']]
        )
        return corrupted_provider

    else:
        raise ValueError(f"Unsupported type of corruptor: {config['type']}")



from typing import List
import torch


def masked_tensor_from_sentences(sentences: List[List[int]], filler=0, device=torch.device('cpu'), target_all=False):
    try:
        sentences[0][0]
    except TypeError:
        raise ValueError("masked_tensor_from_sentences() consumes List of Lists (batch X time)")

    batch_size = len(sentences)
    max_len = max(len(s) for s in sentences)

    shape = (batch_size, max_len-1)
    dtype = torch.int64
    input = torch.zeros(shape, dtype=dtype, device=device)
    target = torch.zeros(shape, dtype=dtype, device=device)
    mask = torch.zeros(shape, dtype=dtype, device=device)

    for s in range(len(sentences)):
        for t in range(len(sentences[s]) - 1):
            input[s, t] = sentences[s][t]
            target[s, t] = sentences[s][t+1]
            mask[s, t] = 1

    if target_all:
        first_inputs = torch.tensor([s[0] for s in sentences], dtype=dtype, device=device)
        target = torch.cat([first_inputs.view(-1, 1), target], dim=1)

        batch_of_ones = torch.ones((batch_size, 1), dtype=mask.dtype, device=mask.device)
        mask = torch.cat([batch_of_ones, mask], dim=1)

    if input.shape[1] == 0:
        batch_of_zeros = torch.zeros((batch_size, 1), dtype=mask.dtype, device=mask.device)
        mask = torch.cat([mask, batch_of_zeros], dim=1)
        input = batch_of_zeros.clone().detach()
        target = torch.cat([target, batch_of_zeros], dim=1)

    return input, target, mask



from brnolm.runtime.tensor_reorganization import TensorReorganizer

import torch
from torch.autograd import Variable
from .common import TestCase


class Dummy_lstm():
    def __init__(self, nb_hidden):
        self._nb_hidden = nb_hidden

    def init_hidden(self, batch_size):
        return (
            torch.FloatTensor([[[0.0] * self._nb_hidden] * batch_size] * 2),
            torch.FloatTensor([[[0.0] * self._nb_hidden] * batch_size] * 2)
        )


class TensorReorganizerTests(TestCase):
    def setUp(self):
        self.lm = Dummy_lstm(nb_hidden=2)
        self.reorganizer = TensorReorganizer(self.lm.init_hidden)

    def test_passing(self):
        last_h = (
            torch.FloatTensor([[[0.1, 0.1], [0.2, 0.2], [0.3, 0.3]]]*2),
            torch.FloatTensor([[[1, 1], [2, 2], [3, 3]]]*2),
        )

        mask = torch.LongTensor([0, 1, 2])
        bsz = 3

        new_h = self.reorganizer(last_h, mask, bsz)
        self.assertEqual(new_h, last_h)

    def test_shrinks(self):
        last_h = (
            torch.FloatTensor([[[0.1, 0.1], [0.2, 0.2], [0.3, 0.3]]]*2),
            torch.FloatTensor([[[1, 1], [2, 2], [3, 3]]]*2),
        )

        mask = torch.LongTensor([0, 2])
        bsz = 2

        new_h = self.reorganizer(last_h, mask, bsz)
        expected = (
            torch.FloatTensor([[[0.1, 0.1], [0.3, 0.3]]]*2),
            torch.FloatTensor([[[1, 1], [3, 3]]]*2),
        )
        self.assertEqual(new_h, expected)

    def test_requires_bsz_greater_than_mask(self):
        last_h = (
            torch.FloatTensor([[[0.1, 0.1], [0.2, 0.2], [0.3, 0.3]]]*2),
            torch.FloatTensor([[[1, 1], [2, 2], [3, 3]]]*2),
        )

        mask = torch.LongTensor([0, 1, 2])
        bsz = 2

        self.assertRaises(ValueError, self.reorganizer, last_h, mask, bsz)

    def test_on_empty_mask_zeros(self):
        last_h = (
            torch.FloatTensor([[[0.1, 0.1], [0.2, 0.2], [0.3, 0.3]]]*2),
            torch.FloatTensor([[[1, 1], [2, 2], [3, 3]]]*2),
        )

        mask = torch.LongTensor([])
        bsz = 2

        new_h = self.reorganizer(last_h, mask, bsz)
        expected = self.lm.init_hidden(bsz)
        self.assertEqual(new_h, expected)

    def test_completion_by_zeros(self):
        last_h = (
            torch.FloatTensor([[[0.1, 0.1], [0.2, 0.2], [0.3, 0.3]]]*2),
            torch.FloatTensor([[[1, 1], [2, 2], [3, 3]]]*2),
        )

        mask = torch.LongTensor([1])
        bsz = 2

        new_h = self.reorganizer(last_h, mask, bsz)
        expected = (
            torch.FloatTensor([[[0.2, 0.2], [0.0, 0.0]]]*2),
            torch.FloatTensor([[[2.0, 2.0], [0.0, 0.0]]]*2),
        )
        self.assertEqual(new_h, expected)

    def test_bug_regression_single_addition(self):
        last_h = (
            torch.FloatTensor([[[0.1, 0.1], [0.2, 0.2], [0.3, 0.3]]]*2),
            torch.FloatTensor([[[1, 1], [2, 2], [3, 3]]]*2),
        )

        mask = torch.LongTensor([1, 2])
        bsz = 3

        new_h = self.reorganizer(last_h, mask, bsz)
        expected = (
            torch.FloatTensor([[[0.2, 0.2], [0.3, 0.3], [0.0, 0.0]]]*2),
            torch.FloatTensor([[[2.0, 2.0], [3.0, 3.0], [0.0, 0.0]]]*2),
        )
        self.assertEqual(new_h, expected)


class Dummy_srn():
    def __init__(self, nb_hidden):
        self._nb_hidden = nb_hidden
        self._nb_layers = 1

    def init_hidden(self, batch_size):
        return torch.FloatTensor(self._nb_layers, batch_size, self._nb_hidden).zero_()


class TensorReorganizerTests_SRN(TestCase):
    def setUp(self):
        lm = Dummy_srn(nb_hidden=2)
        self.reorganizer = TensorReorganizer(lm.init_hidden)

    def test_passing(self):
        last_h = torch.FloatTensor([[[0.1, 0.1], [0.2, 0.2], [0.3, 0.3]]])

        mask = torch.LongTensor([0, 1, 2])
        bsz = 3

        new_h = self.reorganizer(last_h, mask, bsz)
        self.assertEqual(new_h, last_h)

    def test_passing_variables(self):
        last_h = Variable(torch.FloatTensor([[[0.1, 0.1], [0.2, 0.2], [0.3, 0.3]]]))

        mask = Variable(torch.LongTensor([0, 1, 2]))
        bsz = 3

        new_h = self.reorganizer(last_h, mask, bsz)
        self.assertEqual(new_h, last_h)

    def test_shrinks(self):
        last_h = torch.FloatTensor([[[0.1, 0.1], [0.2, 0.2], [0.3, 0.3]]])

        mask = torch.LongTensor([0, 2])
        bsz = 2

        new_h = self.reorganizer(last_h, mask, bsz)
        expected = torch.FloatTensor([[[0.1, 0.1], [0.3, 0.3]]])

        self.assertEqual(new_h, expected)

    def test_requires_bsz_greater_than_mask(self):
        last_h = torch.FloatTensor([[[0.1, 0.1], [0.2, 0.2], [0.3, 0.3]]])

        mask = torch.LongTensor([0, 1, 2])
        bsz = 2

        self.assertRaises(ValueError, self.reorganizer, last_h, mask, bsz)

    def test_on_empty_mask_zeros(self):
        last_h = torch.FloatTensor([[[0.1, 0.1], [0.2, 0.2], [0.3, 0.3]]])

        mask = torch.LongTensor([])
        bsz = 2

        new_h = self.reorganizer(last_h, mask, bsz)
        expected = torch.FloatTensor([[[0.0, 0.0], [0.0, 0.0]]])
        self.assertEqual(new_h, expected)

    def test_completion_by_zeros(self):
        last_h = torch.FloatTensor([[[0.1, 0.1], [0.2, 0.2], [0.3, 0.3]]])

        mask = torch.LongTensor([1])
        bsz = 2

        new_h = self.reorganizer(last_h, mask, bsz)
        expected = torch.FloatTensor([[[0.2, 0.2], [0.0, 0.0]]])
        self.assertEqual(new_h, expected)



import torch


class Substitutor:
    def __init__(self, rate, replacements_range):
        self.rate = rate
        self.replacements_range = replacements_range
        if not isinstance(replacements_range, int) or replacements_range < 0:
            raise ValueError(f"Replacements range needs to be a positive integer, got {replacements_range}")

    def __call__(self, X, targets):
        replacements = torch.randint(0, self.replacements_range, X.shape, device=X.device, dtype=X.dtype)
        mask = torch.full(X.shape, self.rate, device=X.device)
        mask = torch.bernoulli(mask).long()

        return X * (1-mask) + replacements * mask, targets


class Deletor:
    def __init__(self, rate):
        self.rate = rate

    def __call__(self, X, targets):
        timemask = torch.full((X.shape[1], ), 1 - self.rate, device=X.device)
        timemask = torch.bernoulli(timemask).bool()
        return X[:, timemask], targets[:, timemask]


class Corruptor:
    def __init__(self, source, substitution_rate=-1.0, replacements_range=None, deletion_rate=-1.0):
        self.source = source
        if substitution_rate > 0.0:
            self.substitutor = Substitutor(substitution_rate, replacements_range)
        else:
            self.substitutor = None

        if deletion_rate > 0.0:
            self.deletor = Deletor(deletion_rate)
        else:
            self.deletor = None

    def __iter__(self):
        for X, t in self.source:
            if self.deletor:
                X, t = self.deletor(X, t)
            if self.substitutor:
                X, t = self.substitutor(X, t)
            yield X, t



import numpy as np
import copy


def word_distance(a, b):
    return 0 if a == b else 1


DIAGONAL_MOVE = 0
HORIZONAL_MOVE = 1
VERTICAL_MOVE = -1


def path_from_moves(moves_taken):
    ptr_a = moves_taken.shape[0] - 1
    ptr_b = moves_taken.shape[1] - 1

    path = []
    while ptr_a != 0 or ptr_b != 0:
        move = int(moves_taken[ptr_a, ptr_b])
        path.append(move)

        if move == VERTICAL_MOVE:
            ptr_a -= 1
        elif move == HORIZONAL_MOVE:
            ptr_b -= 1
        else:
            ptr_a -= 1
            ptr_b -= 1

    return list(reversed(path))


def ind_ali_from_path(path):
    alignment = []
    ptr_a = 0
    ptr_b = 0
    inds_a = []
    inds_b = []
    for move in path:
        if move == VERTICAL_MOVE:
            inds_a.append(ptr_a)
            ptr_a += 1
        elif move == HORIZONAL_MOVE:
            inds_b.append(ptr_b)
            ptr_b += 1
        else:
            inds_a.append(ptr_a)
            inds_b.append(ptr_b)
            ptr_a += 1
            ptr_b += 1

            alignment.append((inds_a, inds_b))
            inds_a = []
            inds_b = []

    if len(inds_a) > 0 or len(inds_b) > 0:
        assert(len(inds_a) == 0 or len(inds_b) == 0)
        alignment[-1] = (
            alignment[-1][0] + inds_a,
            alignment[-1][1] + inds_b,
        )

    return alignment


def local_costs_from_strings(a, b):
    local_costs = np.zeros(shape=(len(a)+1, len(b)+1))

    for i, w_a in enumerate(a):
        for j, w_b in enumerate(b):
            local_costs[i+1, j+1] = word_distance(w_a, w_b)

    return local_costs


def path_from_local_costs(local_costs):
    partial_costs = np.full(shape=local_costs.shape, fill_value=np.inf)
    moves_taken = np.full(shape=local_costs.shape, fill_value=np.inf)

    for i in range(partial_costs.shape[0]):
        for j in range(partial_costs.shape[1]):
            if i == 0 and j == 0:
                partial_costs[i, j] = 0.0
                moves_taken[i, j] = DIAGONAL_MOVE
            elif i == 0 and j != 0:
                partial_costs[i, j] = partial_costs[i, j-1] + 1.0
                moves_taken[i, j] = HORIZONAL_MOVE
            elif i != 0 and j == 0:
                partial_costs[i, j] = partial_costs[i-1, j] + 1.0
                moves_taken[i, j] = VERTICAL_MOVE
            else:
                vertical_cost = partial_costs[i-1, j] + 1.0
                horizontal_cost = partial_costs[i, j-1] + 1.0
                diagonal_cost = partial_costs[i-1, j-1] + local_costs[i, j]

                best_cost = min([vertical_cost, horizontal_cost, diagonal_cost])

                partial_costs[i, j] = best_cost
                if best_cost == diagonal_cost:
                    moves_taken[i, j] = DIAGONAL_MOVE
                elif best_cost == horizontal_cost:
                    moves_taken[i, j] = HORIZONAL_MOVE
                else:
                    moves_taken[i, j] = VERTICAL_MOVE

    return moves_taken


def word_ali_from_index_ali(a, b, index_alignment):
    word_alignment = []
    for inds_a, inds_b in index_alignment:
        word_alignment.append((
            [a[ind_a] for ind_a in inds_a],
            [b[ind_b] for ind_b in inds_b],
        ))

    return word_alignment


def align(a, b):
    local_costs = local_costs_from_strings(a, b)
    moves_taken = path_from_local_costs(local_costs)
    path = path_from_moves(moves_taken)
    index_alignment = ind_ali_from_path(path)
    return word_ali_from_index_ali(a, b, index_alignment)


def insertion_mismatch(a, b):
    assert(len(a) < len(b))
    assert(len(a) == 1)

    if a[0] == b[0]:
        mismatch = ([], b[1:])
        ends_with_mismatch = True
    elif a[0] == b[-1]:
        mismatch = ([], b[:-1])
        ends_with_mismatch = False
    else:
        mismatch = (a, b)
        ends_with_mismatch = True

    return mismatch, ends_with_mismatch


def equal_lenght_mismatch(a, b):
    assert(len(a) == 1 and len(b) == 1)
    if a != b:
        mismatch = (a, b)
        ends_with_mismatch = True
    else:
        mismatch = None
        ends_with_mismatch = False

    return mismatch, ends_with_mismatch


def single_pair_mismatch(a, b):
    if len(a) == len(b):
        mismatch, ends_with_mismatch = equal_lenght_mismatch(a, b)
    elif len(a) < len(b):
        mismatch, ends_with_mismatch = insertion_mismatch(a, b)
    elif len(a) > len(b):
        mismatch, ends_with_mismatch = insertion_mismatch(b, a)
        mismatch = tuple(reversed(mismatch))

    return mismatch, ends_with_mismatch


def extract_mismatch(ali):
    mismatches = []
    last_was_mismatched = False
    for a, b in ali:
        do_extend = copy.deepcopy(last_was_mismatched)
        mismatch, last_was_mismatched = single_pair_mismatch(a, b)

        if not mismatch:
            continue

        if do_extend:
            mismatches[-1][0].extend(mismatch[0])
            mismatches[-1][1].extend(mismatch[1])
        else:
            mismatches.append(mismatch)

    return mismatches


def find_in_mismatches(mismatches, word):
    for m in mismatches:
        if word in m[0] or word in m[1]:
            return m


def number_of_errors(mismatches):
    nb_errors = 0
    for a, b in mismatches:
        nb_errors += 1 + abs(len(a) - len(b))
    return nb_errors



import random
import torch

from brnolm.data_pipeline import split_corpus_dataset

import sys
import math


class CudaStream():
    def __init__(self, source):
        self._source = source

    def __iter__(self):
        for batch in self._source:
            yield tuple(x.cuda() for x in batch)


class TransposeWrapper:
    def __init__(self, stream):
        self._stream = stream

    def __iter__(self):
        for a_tuple in self._stream:
            yield tuple(x.t().contiguous() for x in a_tuple)

    def __len__(self):
        return len(self._stream)


def repackage_hidden(h):
    """Detaches a tuple of tensors them from their history."""
    if isinstance(h, torch.Tensor):
        return h.detach()
    else:
        return tuple(repackage_hidden(v) for v in h)


def filelist_to_tokenized_splits(filelist_filename, vocab, bptt, wrapper=split_corpus_dataset.TokenizedSplit):
    filenames = filenames_file_to_filenames(filelist_filename)
    tss = []
    for filename in filenames:
        with open(filename, 'r') as f:
            tss.append(wrapper(f, vocab, bptt))

    return tss


def filelist_to_objects(filelist_filename, action):
    filenames = filenames_file_to_filenames(filelist_filename)
    objects = []
    for filename in filenames:
        with open(filename, 'r') as f:
            objects.append(action(f))

    return objects


def filenames_file_to_filenames(filelist_filename):
    with open(filelist_filename) as filelist:
        filenames = filelist.read().split()

    return filenames


def init_seeds(seed, cuda):
    random.seed(seed)
    torch.manual_seed(seed)
    if cuda and torch.cuda.is_available():
        torch.cuda.manual_seed(seed)


class BatchFilter:
    def __init__(self, data, batch_size, bptt, min_batch_size):
        self._data = data
        self._batch_size = batch_size
        self._bptt = bptt
        self._min_batch_size = min_batch_size

        self._nb_skipped_updates = 0
        self._nb_skipped_words = 0
        self._nb_skipped_seqs = 0  # accumulates size of skipped batches

    def __iter__(self):
        for batch in self._data:
            X = batch[0]
            if X.size(0) >= self._min_batch_size:
                yield batch
            else:
                self._nb_skipped_updates += 1
                self._nb_skipped_words += X.size(0) * X.size(1)
                self._nb_skipped_seqs += X.size(0)

    def report(self):
        if self._nb_skipped_updates > 0:
            sys.stderr.write(
                "WARNING: due to skipping, a total of {} updates was skipped,"
                " containing {} words. Avg batch size {}. Equal to {} full batches"
                "\n".format(
                    self._nb_skipped_updates,
                    self._nb_skipped_words,
                    self._nb_skipped_seqs/self._nb_skipped_updates,
                    self._nb_skipped_words/(self._batch_size*self._bptt)
                )
            )


def epoch_summary(epoch_no, nb_updates, elapsed_time, loss):
    delim_line = '-' * 89 + '\n'

    epoch_stmt = 'end of epoch {:3d}'.format(epoch_no)
    updates_stmt = '# updates: {}'.format(nb_updates)
    time_stmt = 'time: {:5.2f}s'.format(elapsed_time)
    loss_stmt = 'valid loss {:5.2f}'.format(loss)
    ppl_stmt = 'valid ppl {:8.2f}'.format(math.exp(loss))
    values_line = '| {} | {} | {} | {} | {}\n'.format(
        epoch_stmt, updates_stmt, time_stmt, loss_stmt, ppl_stmt
    )

    return delim_line + values_line + delim_line



from unittest import TestCase

from brnolm.oov_clustering.oov_alignment_lib import align, extract_mismatch
from brnolm.oov_clustering.oov_alignment_lib import find_in_mismatches, number_of_errors


class AlignTest(TestCase):
    def test_trivial(self):
        a = "a b".split()
        b = "a b".split()

        expected = [
            (['a'], ['a']),
            (['b'], ['b']),
        ]

        self.assertEqual(align(a, b), expected)

    def test_single_element(self):
        a = "a".split()
        b = "a".split()

        expected = [
            (['a'], ['a']),
        ]

        self.assertEqual(align(a, b), expected)

    def test_substitution_only(self):
        a = "a".split()
        b = "b".split()

        expected = [
            (['a'], ['b']),
        ]

        self.assertEqual(align(a, b), expected)

    def test_single_insertion(self):
        a = "a".split()
        b = "a b".split()

        expected = [
            (['a'], ['a', 'b']),
        ]

        self.assertEqual(align(a, b), expected)

    def test_double_insertion(self):
        a = "a".split()
        b = "a b c".split()

        expected = [
            (['a'], ['a', 'b', 'c']),
        ]

        self.assertEqual(align(a, b), expected)

    def test_single_insertion_reversed(self):
        a = "b".split()
        b = "a b".split()

        expected = [
            (['b'], ['a', 'b']),
        ]

        self.assertEqual(align(a, b), expected)

    def test_double_insertion_reversed(self):
        a = "b".split()
        b = "c a b".split()

        expected = [
            (['b'], ['c', 'a', 'b']),
        ]

        self.assertEqual(align(a, b), expected)

    def test_single_deletion(self):
        a = "a b".split()
        b = "a".split()

        expected = [
            (['a', 'b'], ['a']),
        ]

        self.assertEqual(align(a, b), expected)

    def test_double_deletion(self):
        a = "a b c".split()
        b = "a".split()

        expected = [
            (['a', 'b', 'c'], ['a']),
        ]

        self.assertEqual(align(a, b), expected)

    def test_single_deletion_reversed(self):
        a = "a b".split()
        b = "b".split()

        expected = [
            (['a', 'b'], ['b']),
        ]

        self.assertEqual(align(a, b), expected)

    def test_double_deletion_reversed(self):
        a = "c a b".split()
        b = "b".split()

        expected = [
            (['c', 'a', 'b'], ['b']),
        ]

        self.assertEqual(align(a, b), expected)

    def test_double_substituion(self):
        a = "a b c d".split()
        b = "a x y d".split()

        expected = [
            (['a'], ['a']),
            (['b'], ['x']),
            (['c'], ['y']),
            (['d'], ['d']),
        ]

        self.assertEqual(align(a, b), expected)

    def test_inner_insertion(self):
        a = "a d".split()
        b = "a b c d".split()

        expected_1 = [
            (['a'], ['a', 'b', 'c']),
            (['d'], ['d']),
        ]
        expected_2 = [
            (['a'], ['a']),
            (['d'], ['b', 'c', 'd']),
        ]

        self.assertIn(align(a, b), [expected_1, expected_2])

    def test_inversion_at_end(self):
        a = "a b c d".split()
        b = "a b d c".split()

        expected = [
            (['a'], ['a']),
            (['b'], ['b']),
            (['c'], ['d']),
            (['d'], ['c']),
        ]

        self.assertEqual(align(a, b), expected)


class MismatchExtractionTest(TestCase):
    def test_trivial(self):
        ali = [
            (['a'], ['a'])
        ]
        expectation = [
        ]

        self.assertEqual(extract_mismatch(ali), expectation)

    def test_single_substitution(self):
        ali = [
            (['a'], ['b'])
        ]
        expectation = [
            (['a'], ['b'])
        ]

        self.assertEqual(extract_mismatch(ali), expectation)

    def test_double_substitution(self):
        ali = [
            (['a'], ['a']),
            (['a'], ['b']),
            (['a'], ['b']),
            (['a'], ['a']),
            (['a'], ['b']),
        ]
        expectation = [
            (['a', 'a'], ['b', 'b']),
            (['a'], ['b'])
        ]

        self.assertEqual(extract_mismatch(ali), expectation)

    def test_substitution_with_insertion(self):
        ali = [
            (['begin'], ['begin']),
            (['c'], ['a', 'b']),
            (['end'], ['end']),
        ]
        expectation = [
            (['c'], ['a', 'b']),
        ]

        self.assertEqual(extract_mismatch(ali), expectation)

    def test_insertion_only(self):
        ali = [
            (['begin'], ['begin']),
            (['a'], ['a', 'b']),
            (['end'], ['end']),
        ]
        expectation = [
            ([], ['b']),
        ]

        self.assertEqual(extract_mismatch(ali), expectation)

    def test_insertion_only_reversed(self):
        ali = [
            (['begin'], ['begin']),
            (['b'], ['a', 'b']),
            (['end'], ['end']),
        ]
        expectation = [
            ([], ['a']),
        ]

        self.assertEqual(extract_mismatch(ali), expectation)

    def test_substitution_with_deletion(self):
        ali = [
            (['begin'], ['begin']),
            (['a', 'b'], ['c']),
            (['end'], ['end']),
        ]
        expectation = [
            (['a', 'b'], ['c']),
        ]

        self.assertEqual(extract_mismatch(ali), expectation)

    def test_deletion_only(self):
        ali = [
            (['begin'], ['begin']),
            (['a', 'b'], ['a']),
            (['end'], ['end']),
        ]
        expectation = [
            (['b'], []),
        ]

        self.assertEqual(extract_mismatch(ali), expectation)

    def test_deletion_only_reversed(self):
        ali = [
            (['begin'], ['begin']),
            (['a', 'b'], ['b']),
            (['end'], ['end']),
        ]
        expectation = [
            (['a'], []),
        ]

        self.assertEqual(extract_mismatch(ali), expectation)

    def test_deletion_after_substitution(self):
        ali = [
            (['begin'], ['begin']),
            (['x'], ['y']),
            (['a', 'b'], ['b']),
            (['end'], ['end']),
        ]
        expectation = [
            (['x', 'a'], ['y']),
        ]

        self.assertEqual(extract_mismatch(ali), expectation)


class FindingInMismatchesTest(TestCase):
    def test_trivial(self):
        mismatches = [
            (['x'], ['a'])
        ]
        expectation = (['x'], ['a'])

        self.assertEqual(find_in_mismatches(mismatches, 'x'), expectation)

    def test_finding(self):
        mismatches = [
            (['c'], ['d']),
            (['x'], ['a']),
            (['b'], ['f']),
        ]
        expectation = (['x'], ['a'])

        self.assertEqual(find_in_mismatches(mismatches, 'x'), expectation)

    def test_multiple_elements_in_mismatch(self):
        mismatches = [
            (['c'], ['d']),
            (['c', 'b', 'x'], ['a']),
            (['b'], ['f']),
        ]
        expectation = (['c', 'b', 'x'], ['a'])

        self.assertEqual(find_in_mismatches(mismatches, 'x'), expectation)


class NumberOfErrorsTest(TestCase):
    def test_trivial(self):
        mismatches = [
            (['x'], ['a'])
        ]

        self.assertEqual(number_of_errors(mismatches), 1)

    def test_no_mismatch(self):
        self.assertEqual(number_of_errors([]), 0)

    def test_multiple_substitutions(self):
        mismatches = [
            (['x'], ['a']),
            (['x'], ['a'])
        ]

        self.assertEqual(number_of_errors(mismatches), 2)

    def test_single_insertion(self):
        mismatches = [
            (['x'], ['a', 'b']),
        ]

        self.assertEqual(number_of_errors(mismatches), 2)

    def test_single_deletion(self):
        mismatches = [
            (['a', 'b'], ['x']),
        ]

        self.assertEqual(number_of_errors(mismatches), 2)



import torch
import torch.nn as nn
import torch.nn.functional as F


class BengioModel(nn.Module):
    """Container module with an encoder, a recurrent module, and a decoder."""

    def __init__(self, ntoken, emb_size, in_len, nb_hidden, dropout=0.5):
        super().__init__()
        self.drop = nn.Dropout(dropout)
        self.encoder = nn.Embedding(ntoken, emb_size)
        self.emb2h = nn.ModuleList([nn.Linear(emb_size, nb_hidden) for _ in range(in_len)])

        self.init_weights()

        self.nb_hidden = nb_hidden
        self.in_len = in_len
        self.emb_size = emb_size

        self.batch_first = True

    def init_weights(self):
        initrange = 0.1
        self.encoder.weight.data.uniform_(-initrange, initrange)
        for e2h in self.emb2h:
            e2h.weight.data.uniform_(-initrange, initrange)

    def forward(self, input, hidden):
        emb = self.drop(self.encoder(input))
        projections = [proj(emb[:, i:emb.size(1)-(self.in_len-i)+1]) for i, proj in enumerate(self.emb2h)]
        projections = torch.stack(projections, dim=-1)
        output = F.tanh(torch.sum(projections, dim=-1))
        output = self.drop(output)
        return output, hidden

    def init_hidden(self, bsz):
        # not used, but to fit into the framework of other ivec-LMs
        weight = next(self.parameters()).data
        return weight.new_zeros(1, bsz, self.nb_hidden)


class BengioModelIvecInput(nn.Module):
    """Container module with an encoder, a recurrent module, and a decoder."""

    def __init__(self, ntoken, emb_size, in_len, nb_hidden, dropout, ivec_dim):
        super().__init__()
        self.drop = nn.Dropout(dropout)
        self.encoder = nn.Embedding(ntoken, emb_size)
        self.emb2h = nn.ModuleList([nn.Linear(emb_size, nb_hidden) for _ in range(in_len)])
        self.ivec2h = nn.Linear(ivec_dim, nb_hidden)

        self.init_weights()

        self.nb_hidden = nb_hidden
        self.in_len = in_len
        self.emb_size = emb_size

        self.batch_first = True

    def init_weights(self):
        initrange = 0.1
        self.encoder.weight.data.uniform_(-initrange, initrange)
        for e2h in self.emb2h:
            e2h.weight.data.uniform_(-initrange, initrange)

    def forward(self, input, hidden, ivec):
        if len(ivec.size()) == 1:
            ivec = ivec.unsqueeze(0)
        emb = self.drop(self.encoder(input))
        projections = [proj(emb[:, i:emb.size(1)-(self.in_len-i)+1]) for i, proj in enumerate(self.emb2h)]
        nb_timesteps = projections[0].size(1)
        projected_ivec = self.ivec2h(ivec).unsqueeze(dim=-2).expand(-1, nb_timesteps, -1)
        projections = torch.stack(projections + [projected_ivec], dim=-1)
        output = F.tanh(torch.sum(projections, dim=-1))
        output = self.drop(output)
        return output, hidden

    def init_hidden(self, bsz):
        # not used, but to fit into the framework of other ivec-LMs
        weight = next(self.parameters()).data
        return weight.new_zeros(1, bsz, self.nb_hidden)



import threading
import queue


class DataCreator(threading.Thread):
    def __init__(self, q, data_stream, device):
        super().__init__(daemon=True)
        self.q = q
        self.in_stream = data_stream
        self.device = device

    def run(self):
        for batch in self.in_stream:
            batch = (x.to(self.device) for x in batch)

            self.q.put(batch)


class OndemandDataProvider:
    def __init__(self, in_data, device):
        self.data = in_data
        self.device = device

    def __iter__(self):
        q = queue.Queue(maxsize=10)
        feeder_thread = DataCreator(q, self.data, self.device)
        feeder_thread.start()

        while feeder_thread.is_alive() or not q.empty():
            yield q.get()



#!/usr/bin/env python3

from setuptools import setup


def get_long_desc():
    with open('README.md') as f:
        return f.read()


setup(
    name='brnolm',
    version='0.2.0',
    python_requires='>=3.6',
    packages=[
        'brnolm',
        'brnolm/data_pipeline',
        'brnolm/language_models',
        'brnolm/oov_clustering',
        'brnolm/runtime',
        'brnolm/smm_itf',
    ],
    install_requires=[
        'torch>=1.4',
        'numpy',
        'scikit-learn',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    url='https://github.com/BUTSpeechFIT/BrnoLM',
    description='A language modeling toolkit',
    long_description=get_long_desc(),
    long_description_content_type='text/markdown',
    author='Karel Benes',
    author_email='ibenes@fit.vutbr.cz',
)



#!/usr/bin/env python

import pickle
import argparse

import plotting

import numpy as np


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("--norms", action='store_true',
        help="print average norm of the gradient per layer")
    parser.add_argument("grads")
    args = parser.parse_args()

    with open(args.grads, 'rb') as f:
        grads = pickle.load(f)

    fig_titles = []
    if args.norms:
        for g in grads:
            norms = np.linalg.norm(g, ord=2, axis=1)
            fig_titles.append(", ".join([str(x) for x in [np.min(norms), np.mean(norms), np.max(norms)]]))

    coloring = {
        'vmin' : -5e-3,
        'vmax' : 5e-3,
        'cmap' : 'RdBu'
    }

    plotting.grid_plot(grads, lambda x:x, "Weighted grads", fig_titles, coloring)



#!/usr/bin/env python
import argparse
import io
import math
import sys
import torch
import numpy as np

import split_corpus_dataset
import ivec_appenders
import smm_ivec_extractor

from runtime_utils import filenames_file_to_filenames

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--file-list', required=True,
                        help="file with list of files analyze")
    parser.add_argument('--ivec-extractor', required=True,
                        help="iVector extractor to use")
    parser.add_argument('--output', required=True,
                        help="where to put the ivectors")
    args = parser.parse_args()
    print(args)

    print("loading SMM iVector extractor ...")
    with open(args.ivec_extractor, 'rb') as f:
        ivec_extractor = smm_ivec_extractor.load(f)
    print(ivec_extractor)

    documents = filenames_file_to_filenames(args.file_list)

    ivecs = []
    for doc in documents:
        with open(doc) as f:
            content = f.read()

        complete_ivec = ivec_extractor(content)
        ivecs.append(complete_ivec)

    ivecs = torch.stack(ivecs)
    print(ivecs)

    with open(args.output, 'w') as f:
        np.savetxt(f, ivecs.cpu().numpy()) 



#!/usr/bin/env python

import pickle
import argparse

import plotting

import numpy as np


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("--color-bar", action='store_true')
    parser.add_argument("activations")
    args = parser.parse_args()

    with open(args.activations, 'rb') as f:
        activations = pickle.load(f)

    fig_titles = []
    for a in activations:
        nb_zeros = np.sum(a==0.0)
        fig_titles.append(", ".join([str(x) for x in [nb_zeros, np.mean(a), np.max(a)]]))

    color_setup = {
        'cmap': 'gnuplot',
        'vmin': 0,
        'vmax': 30,
        'colorbar': args.color_bar,
    }

    plotting.grid_plot(activations, lambda x:x, "activations", fig_titles, coloring=color_setup)
